from __future__ import annotations

import fnmatch
import hashlib
import json
import re
import shutil
import threading
import warnings
import weakref

from concurrent import futures
from functools import wraps
from itertools import product
from os.path import basename
from pathlib import Path
from typing import (
    TYPE_CHECKING, Any, ClassVar, Literal,
)
from urllib.parse import quote, urlparse

import numpy as np
import pandas as pd
import panel as pn
import param  # type: ignore
import requests

from panel.io.cache import _generate_hash

from ..base import MultiTypeComponent
from ..filters.base import Filter
from ..state import state
from ..transforms.base import Filter as FilterTransform, Transform
from ..transforms.sql import (
    SQLCount, SQLDistinct, SQLLimit, SQLMinMax, SQLSample, SQLSelectFrom,
    SQLTransform,
)
from ..util import get_dataframe_schema, is_ref, merge_schemas
from ..validation import ValidationError, match_suggestion_message

if TYPE_CHECKING:
    from dask.dataframe import DataFrame as dDataFrame, Series as dSeries
    from panel.viewable import Viewable

    DataFrame = pd.DataFrame | dDataFrame
    Series = pd.Series | dSeries

    DataFrameTypes: tuple[type, ...]
    try:
        import dask.dataframe as dd
        DataFrameTypes = (pd.DataFrame, dd.DataFrame)
    except ImportError:
        dd = None  # type: ignore
        DataFrameTypes = (pd.DataFrame,)



def cached(method, locks=weakref.WeakKeyDictionary()):
    """
    Adds caching to a Source.get query.

    Returns
    -------
    Returns method wrapped in caching functionality.
    """
    @wraps(method)
    def wrapped(self, table, **query):
        if self._supports_sql and not self.cache_per_query and 'sql_transforms' in query:
            raise RuntimeError(
                'SQLTransforms cannot be used on a Source with cache_per_query '
                'being disabled. Ensure you set `cache_per_query=True`.'
            )
        if self in locks:
            main_lock = locks[self]['main']
        else:
            main_lock = threading.RLock()
            locks[self] = {'main': main_lock}
        with main_lock:
            if table in locks:
                lock = locks[self][table]
            else:
                locks[self][table] = lock = threading.RLock()
        cache_query = query if self.cache_per_query else {}
        with lock:
            df, no_query = self._get_cache(table, **cache_query)
        if df is None:
            if not self.cache_per_query and (hasattr(self, 'dask') or hasattr(self, 'use_dask')):
                cache_query['__dask'] = True
            df = method(self, table, **cache_query)
            with lock:
                self._set_cache(df, table, **cache_query)
        filtered = df
        if (not self.cache_per_query or no_query) and query:
            filtered = FilterTransform.apply_to(
                df, conditions=list(query.items())
            )
        if getattr(self, 'dask', False) or not hasattr(filtered, 'compute'):
            return filtered
        return filtered.compute()
    return wrapped


def cached_schema(method, locks=weakref.WeakKeyDictionary()):
    @wraps(method)
    def wrapped(self, table: str | None = None, limit: int | None = None, shuffle: bool = False):
        if self in locks:
            main_lock = locks[self]['main']
        else:
            main_lock = threading.RLock()
            locks[self] = {'main': main_lock}
        with main_lock:
            schema = self._get_schema_cache() or {}
        tables = self.get_tables() if table is None else [table]
        if all(table in schema for table in tables) and limit is None and not shuffle:
            return schema if table is None else schema[table]
        for missing in tables:
            if missing in schema:
                continue
            with main_lock:
                if missing in locks[self]:
                    lock = locks[self][missing]
                else:
                    locks[self][missing] = lock = threading.RLock()
            with lock:
                with main_lock:
                    new_schema = self._get_schema_cache() or {}
                if missing in new_schema and limit is None and not shuffle:
                    schema[missing] = new_schema[missing]
                else:
                    schema[missing] = method(self, missing, limit, shuffle)
            with main_lock:
                self._set_schema_cache(schema)
        return schema if table is None else schema[table]
    return wrapped


def cached_metadata(method, locks=weakref.WeakKeyDictionary()):
    @wraps(method)
    def wrapped(self, table: str | list[str] | None = None):
        if self in locks:
            main_lock = locks[self]['main']
        else:
            main_lock = threading.RLock()
            locks[self] = {'main': main_lock}
        with main_lock:
            metadata = self._get_metadata_cache() or {}
        if table is None:
            tables = self.get_tables()
        elif isinstance(table, str):
            tables = [table]
        else:
            tables = table
        if all(t in metadata for t in tables):
            if isinstance(table, str):
                return metadata[table]
            return {table: metadata[table] for table in tables}

        missing_tables = [t for t in tables if t not in metadata]
        metadata.update(method(self, missing_tables))
        with main_lock:
            self._set_metadata_cache(metadata)
        if isinstance(table, str):
            return metadata[table]
        return {table: metadata[table] for table in tables}
    return wrapped


class Source(MultiTypeComponent):
    """
    `Source` components provide allow querying all kinds of data.

    A `Source` can return one or more tables queried using the
    `.get_tables` method, a description of the data returned by each
    table in the form of a JSON schema accessible via the `.get_schema`
    method and lastly a `.get` method that allows filtering the data.

    The Source base class also implements both in-memory and disk
    caching which can be enabled if a `cache_dir` is provided. Data
    cached to disk is stored as parquet files.
    """

    cache_with_dask = param.Boolean(default=True, doc="""
        Whether to read and write cache files with dask if available.""")

    cache_per_query = param.Boolean(default=True, doc="""
        Whether to query the whole dataset or individual queries.""")

    cache_dir = param.String(default=None, doc="""
        Whether to enable local cache and write file to disk.""")

    cache_data = param.Boolean(default=True, doc="""
        Whether to cache actual data.""")

    cache_schema = param.Boolean(default=True, doc="""
        Whether to cache table schemas.""")

    cache_metadata = param.Boolean(default=True, doc="""
        Whether to cache metadata.""")

    metadata_func = param.Callable(default=None, doc="""
        Function to implement custom metadata lookup for tables.
        Given a list of tables it should return a dictionary of the form:

        {
            <table>: {"description": ..., "columns": {"column_name": "..."}}
        }

        May be used to override the default _get_table_metadata
        implementation of the Source.""")

    root = param.ClassSelector(class_=Path, precedence=-1, doc="""
        Root folder of the cache_dir, default is config.root""")

    shared = param.Boolean(default=False, doc="""
        Whether the Source can be shared across all instances of the
        dashboard. If set to `True` the Source will be loaded on
        initial server load.""")

    source_type: ClassVar[str | None] = None

    __abstract = True

    # Specification configuration
    _internal_params: ClassVar[list[str]] = ['name', 'root']

    # Declare whether source supports SQL transforms
    _supports_sql: ClassVar[bool] = False

    # Valid keys incude all parameters (except _internal_params)
    _valid_keys: ClassVar[list[str] | Literal['params'] | None] = 'params'

    @property
    def _reload_params(self) -> list[str]:
        "List of parameters that trigger a data reload."
        return list(self.param)

    @classmethod
    def _recursive_resolve(
        cls, spec: dict[str, Any], source_type: type[Source]
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        resolved_spec, refs = {}, {}
        if 'sources' in source_type.param and 'sources' in spec:
            resolved_spec['sources'] = {
                source: cls.from_spec(source)
                for source in spec.pop('sources')
            }
        if 'source' in source_type.param and 'source' in spec:
            resolved_spec['source'] = cls.from_spec(spec.pop('source'))
        for k, v in spec.items():
            if is_ref(v):
                refs[k] = v
                v = state.resolve_reference(v)
            elif isinstance(v, dict):
                v, subrefs = cls._recursive_resolve(v, source_type)
                for sk, sv in subrefs.items():
                    refs[f'{k}.{sk}'] = sv
            if k == 'filters' and 'source' in resolved_spec:
                assert isinstance(resolved_spec['source'], Source)
                source_schema = resolved_spec['source'].get_schema()
                v = [Filter.from_spec(fspec, source_schema) for fspec in v]
            if k == 'transforms':
                v = [Transform.from_spec(tspec) for tspec in v]
            resolved_spec[k] = v
        return resolved_spec, refs

    @classmethod
    def _validate_filters(
        cls, filter_specs: dict[str, dict[str, Any] | str], spec: dict[str, Any], context: dict[str, Any]
    ) -> dict[str, Any]:
        warnings.warn(
            'Providing filters in a Source definition is deprecated, '
            'please declare filters as part of a Pipeline.', DeprecationWarning
        )
        return cls._validate_dict_subtypes('filters', Filter, filter_specs, spec, context)

    @classmethod
    def validate(cls, spec: dict[str, Any] | str, context: dict[str, Any] | None = None) -> dict[str, Any] | str:
        if isinstance(spec, str):
            if context is None or spec not in context.get('sources', {}):
                msg = f'Referenced non-existent source {spec!r}.'
                sources = list(context.get('sources', {})) if context else []
                msg = match_suggestion_message(spec, sources, msg)
                raise ValidationError(msg, spec, spec)
            return spec
        return super().validate(spec, context)

    @classmethod
    def from_spec(cls, spec: dict[str, Any] | str) -> Source:
        """
        Creates a Source object from a specification. If a Source
        specification references other sources these may be supplied
        in the sources dictionary and be referenced by name.

        Parameters
        ----------
        spec : dict or str
            Specification declared as a dictionary of parameter values
            or a string referencing a source in the sources dictionary.

        Returns
        -------
        Resolved and instantiated Source object
        """
        if isinstance(spec, str):
            if spec in state.sources:
                source = state.sources[spec]
            elif spec in state.spec.get('sources', {}):
                source = state.load_source(spec, state.spec['sources'][spec])
            return source

        spec = spec.copy()
        src_type_name = spec.pop('type', None)
        source_type = Source._get_type(src_type_name)
        if cls is Source:
            spec['type'] = src_type_name
            return source_type.from_spec(spec)
        resolved_spec, refs = source_type._recursive_resolve(spec, source_type)
        return source_type(refs=refs, **resolved_spec)

    def __init__(self, **params):
        from ..config import config
        params['root'] = Path(params.get('root', config.root))
        super().__init__(**params)
        self.param.watch(self.clear_cache, self._reload_params)
        self._cache = {}
        self._schema_cache = {}
        self._metadata_cache = {}

    def _get_key(self, table: str, **query) -> str:
        sha = hashlib.sha256()
        sha.update(self._get_source_hash().encode('utf-8'))
        sha.update(table.encode('utf-8'))
        if 'sql_transforms' in query:
            sha.update(_generate_hash([hash(t) for t in query.pop('sql_transforms')]))
        sha.update(_generate_hash(query))
        return sha.hexdigest()

    def _get_source_hash(self):
        sha = hashlib.sha256()
        for k, v in self.param.values().items():
            if k in ('root',):
                continue
            sha.update(k.encode('utf-8'))
            sha.update(_generate_hash(v))
        return sha.hexdigest()

    def _get_metadata_cache(self) -> dict[str, dict[str, Any]]:
        metadata = self._metadata_cache if self._metadata_cache else None
        sha = self._get_source_hash()
        if self.cache_dir:
            path = self.root / self.cache_dir / f'{self.name}_{sha}_metadata.json'
            if not path.is_file():
                return metadata
            with open(path) as f:
                json_metadata = json.load(f)
            if metadata is None:
                metadata = {}
            for table, table_metadata in json_metadata.items():
                if table not in metadata:
                    metadata[table] = table_metadata
        return metadata

    def _get_schema_cache(self) -> dict[str, dict[str, Any]]:
        schema = self._schema_cache if self._schema_cache else None
        sha = self._get_source_hash()
        if self.cache_dir:
            path = self.root / self.cache_dir / f'{self.name}_{sha}.json'
            if not path.is_file():
                return schema
            with open(path) as f:
                json_schema = json.load(f)
            if schema is None:
                schema = {}
            for table, tschema in json_schema.items():
                if table in schema:
                    continue
                for col, cschema in tschema.items():
                    if isinstance(cschema, int):
                        continue
                    if cschema.get('type') == 'string' and cschema.get('format') == 'datetime':
                        cschema['inclusiveMinimum'] = pd.to_datetime(
                            cschema['inclusiveMinimum']
                        )
                        cschema['inclusiveMaximum'] = pd.to_datetime(
                            cschema['inclusiveMaximum']
                        )
                schema[table] = tschema
        return schema

    def _set_metadata_cache(self, metadata):
        if not self.cache_metadata:
            return
        self._metadata_cache = metadata
        if not self.cache_dir:
            return
        sha = self._get_source_hash()
        path = self.root / self.cache_dir
        path.mkdir(parents=True, exist_ok=True)
        try:
            with open(path / f'{self.name}_{sha}_metadata.json', 'w') as f:
                json.dump(metadata, f, default=str)
        except Exception as e:
            self.param.warning(
                f"Could not cache metadata to disk. Error while "
                f"serializing metadata: {e}"
            )

    def _set_schema_cache(self, schema):
        if not self.cache_metadata:
            return
        self._schema_cache = schema
        if not self.cache_dir:
            return
        sha = self._get_source_hash()
        path = self.root / self.cache_dir
        path.mkdir(parents=True, exist_ok=True)
        try:
            with open(path / f'{self.name}_{sha}.json', 'w') as f:
                json.dump(schema, f, default=str)
        except Exception as e:
            self.param.warning(
                f"Could not cache schema to disk. Error while "
                f"serializing schema: {e}"
            )

    def _get_cache(self, table: str, **query) -> tuple[DataFrame | None, bool]:
        query.pop('__dask', None)
        key = self._get_key(table, **query)
        if key in self._cache:
            return self._cache[key], not bool(query)
        elif self.cache_dir:
            if self.cache_with_dask:
                try:
                    import dask.dataframe as dd
                except Exception:
                    dd = None
            else:
                dd = None
            if query:
                filename = f'{key}_{table}.parq'
            else:
                filename = f'{table}.parq'
            path = self.root / self.cache_dir / filename
            if path.is_file():
                return pd.read_parquet(path), not bool(query)
            if dd and path.is_dir():
                return dd.read_parquet(path), not bool(query)
            path = path.with_suffix('')
            if dd and path.is_dir():
                return dd.read_parquet(path), not bool(query)
        return None, not bool(query)

    def _set_cache(
        self, data: DataFrame, table: str, write_to_file: bool = True, **query
    ):
        if not self.cache_data:
            return
        query.pop('__dask', None)
        key = self._get_key(table, **query)
        self._cache[key] = data
        if self.cache_dir and write_to_file:
            if self.cache_with_dask:
                try:
                    import dask.dataframe as dd
                except Exception:
                    dd = None
            else:
                dd = None
            path = self.root / self.cache_dir
            path.mkdir(parents=True, exist_ok=True)
            if query:
                filename = f'{key}_{table}.parq'
            else:
                filename = f'{table}.parq'
            filepath = path / filename
            if dd:
                if isinstance(data, dd.DataFrame):
                    filepath = filepath.with_suffix('')
            try:
                data.to_parquet(filepath)
            except Exception as e:
                if path.is_file():
                    path.unlink()
                elif path.is_dir():
                    shutil.rmtree(path)
                self.param.warning(
                    f"Could not cache '{table}' to parquet file. "
                    f"Error during saving process: {e}"
                )

    def _get_table_metadata(self, tables: list[str]) -> dict:
        return {}

    def __contains__(self, table):
        return table in self.get_tables()

    def clear_cache(self, *events: param.parameterized.Event):
        """
        Clears any cached data.
        """
        self._cache = {}
        self._schema_cache = {}
        self._metadata_cache = {}
        if self.cache_dir:
            path = self.root / self.cache_dir
            if path.is_dir():
                shutil.rmtree(path)

    @property
    def panel(self) -> Viewable | None:
        """
        A Source can return a Panel object which displays information
        about the Source or controls how the Source queries data.
        """
        return None

    def get_tables(self) -> list[str]:
        """
        Returns the list of tables available on this source.

        Returns
        -------
        list
            The list of available tables on this source.
        """

    @cached_schema
    def get_schema(
        self, table: str | None = None, limit: int | None = None, shuffle: bool = False
    ) -> dict[str, dict[str, Any]] | dict[str, Any]:
        """
        Returns JSON schema describing the tables returned by the
        Source.

        Parameters
        ----------
        table : str | None
            The name of the table to return the schema for. If None
            returns schema for all available tables.
        limit : int | None
            Limits the number of rows considered for the schema calculation

        Returns
        -------
        dict
            JSON schema(s) for one or all the tables.
        """
        schemas = {}
        names = list(self.get_tables())
        for name in names:
            if table is not None and name != table:
                continue
            df = self.get(name, __dask=True)
            schemas[name] = get_dataframe_schema(df)['items']['properties']
        try:
            return schemas if table is None else schemas[table]
        except KeyError as e:
            msg = f"{type(self).name} does not contain '{table}'"
            msg = match_suggestion_message(table or '', names, msg)
            raise ValidationError(msg) from e

    @cached_metadata
    def get_metadata(self, table: str | list[str] | None) -> dict:
        """
        Returns metadata for one, multiple or all tables provided by the source.

        The metadata for a table is structured as:

        {
            "description": ...,
            "columns": {
                <COLUMN>: {
                   "description": ...,
                   "data_type": ...,
                }
            },
            **other_metadata
        }

        If a list of tables or no table is provided the metadata is nested one additional level:

        {
            "table_name": {
                {
                    "description": ...,
                    "columns": {
                        <COLUMN>: {
                        "description": ...,
                        "data_type": ...,
                        }
                    },
                    **other_metadata
                }
            }
        }

        Parameters
        ----------
        table : str | list[str] | None
            The name of the table to return the schema for. If None
            returns schema for all available tables.

        Returns
        -------
        metadata : dict
            Dictionary of metadata indexed by table (if no table was
            was provided or individual table metdata.
        """
        if table is None:
            tables = self.get_tables()
        elif isinstance(table, str):
            tables = [table]
        else:
            tables = table
        if self.metadata_func:
            metadata = self.metadata_func(tables)
        else:
            metadata = self._get_table_metadata(tables)
        return metadata

    def get(self, table: str, **query) -> DataFrame:
        """
        Return a table; optionally filtered by the given query.

        Parameters
        ----------
        table : str
             The name of the table to query
        query : dict
             A dictionary containing all the query parameters

        Returns
        -------
        DataFrame
            A DataFrame containing the queried table.
        """

    def __str__(self) -> str:
        return self.name


class RESTSource(Source):
    """
    `RESTSource` allows querying REST endpoints conforming to the Lumen REST specification.

    The `url` must offer two endpoints, the `/data` endpoint must
    return data in a records format while the `/schema` endpoint must
    return a valid Lumen JSON schema.
    """

    url = param.String(doc="URL of the REST endpoint to monitor.")

    source_type: ClassVar[str] = 'rest'

    @cached_schema
    def get_schema(
        self, table: str | None = None, limit: int | None = None, shuffle: bool = False
    ) -> dict[str, dict[str, Any]] | dict[str, Any]:
        query = {} if table is None else {'table': table}
        response = requests.get(self.url+'/schema', params=query)
        return {table: schema['items']['properties'] for table, schema in
                response.json().items()}

    @cached
    def get(self, table: str, **query) -> pd.DataFrame:
        query = dict(table=table, **query)
        r = requests.get(self.url+'/data', params=query)
        df = pd.DataFrame(r.json())
        return df


class InMemorySource(Source):
    """
    `InMemorySource` can be used to work with in-memory data.
    """

    tables = param.Dict(default={})

    def get_tables(self) -> list[str]:
        return list(self.tables)

    def get_schema(
        self, table: str | None = None, limit: int | None = None, shuffle: bool = False
    ) -> dict[str, dict[str, Any]] | dict[str, Any]:
        if table:
            df = self.get(table)
            return get_dataframe_schema(df)['items']['properties']
        else:
            return {t: get_dataframe_schema(self.get(t))['items']['properties'] for t in self.get_tables()}

    def get(self, table: str, **query) -> pd.DataFrame:
        dask = query.pop('__dask', False)
        table = self.tables.get(table)
        df = FilterTransform.apply_to(table, conditions=list(query.items()))
        return df if dask or not hasattr(df, 'compute') else df.compute()

    def add_table(self, name, table):
        self.tables[name] = table


class FileSource(Source):
    """
    `FileSource` loads CSV, Excel and Parquet files using pandas and dask `read_*` functions.

    The `FileSource` can declare a list or dictionary of local or
    remote files which are then loaded using either `pandas.read_*` or
    `dask.dataframe.read_*` functions depending on whether `use_dask`
    is enabled.
    """

    dask = param.Boolean(default=False, doc="""
        Whether to return a Dask dataframe.""")

    kwargs = param.Dict(doc="""
        Keyword arguments to the pandas/dask loading function.""")

    tables = param.ClassSelector(class_=(list, dict), doc="""
        List or dictionary of tables to load. If a list is supplied the
        names are computed from the filenames, otherwise the keys are
        the names. The values must filepaths or URLs to the data:

        ```
        {
            'local' : '/home/user/local_file.csv',
            'remote': 'https://test.com/test.csv'
        }
        ```

        if the filepath does not have a declared extension an extension
        may be provided in a list or tuple, e.g.:

        ```
        {'table': ['http://test.com/api', 'json']}
        ```
        """)

    use_dask = param.Boolean(default=True, doc="""
        Whether to use dask to load files.""")

    _pd_load_fns = {
        'csv': pd.read_csv,
        'xlsx': pd.read_excel,
        'xls': pd.read_excel,
        'parq': pd.read_parquet,
        'parquet': pd.read_parquet,
        'json': pd.read_json
    }

    _load_kwargs: ClassVar[dict[str, dict[str, Any]]] = {
        'csv': {'parse_dates': True}
    }

    _template_re = re.compile(r'(\$\{[\w.]+\})')

    source_type: ClassVar[str] = 'file'

    def __init__(self, **params):
        if 'files' in params:
            params['tables'] = params.pop('files')
        super().__init__(**params)

    def _load_fn(self, ext, dask=True):
        kwargs = dict(self._load_kwargs.get(ext, {}))
        if self.kwargs:
            kwargs.update(self.kwargs)
        if self.use_dask and dask:
            try:
                import dask.dataframe as dd
            except Exception:
                return self._load_fn(ext, dask=False)
            if ext == 'csv':
                return dd.read_csv, kwargs
            elif ext in ('parq', 'parquet'):
                return dd.read_parquet, kwargs
            elif ext == 'json':
                if 'orient' not in kwargs:
                    kwargs['orient'] = None
                return dd.read_json, kwargs
        if ext not in self._pd_load_fns:
            raise ValueError(f"File type '{ext}' not recognized and cannot be loaded.")
        return self._pd_load_fns[ext], kwargs

    def _set_cache(self, data: DataFrame, table: str, write_to_file: bool = True, **query):
        file, ext = self._named_files[table]
        if ext in ('parq', 'parquet') and Path(file).exists():
            write_to_file = False
        super()._set_cache(data, table, write_to_file, **query)

    @property
    def _named_files(self) -> dict[str, tuple[str, str]]:
        if isinstance(self.tables, list):
            tables = {}
            for f in self.tables:
                if isinstance(f, str) and f.startswith('http'):
                    name = f
                else:
                    name = '.'.join(basename(f).split('.')[:-1])
                tables[name] = f
        else:
            tables = self.tables or {}
        files = {}
        for name, table in tables.items():
            if isinstance(table, (list | tuple)):
                table, ext = table
            else:
                if isinstance(table, str) and table.startswith('http'):
                    file = basename(urlparse(table).path)
                else:
                    file = basename(table)
                ext = re.search(r"\.(\w+)$", file)
                if ext:
                    ext = ext.group(1)
            files[name] = (table, ext)
        return files

    def _resolve_template_vars(self, table: str) -> list[str]:
        for m in self._template_re.findall(str(table)):
            values = state.resolve_reference(f'${m[2:-1]}')
            values = ','.join([v for v in values])
            table = table.replace(m, quote(values))
        return [table]

    def get_tables(self) -> list[str]:
        return list(self._named_files)

    def _load_table(self, table: str, dask: bool = True) -> DataFrame:
        df = None
        for name, (filepath, ext) in self._named_files.items():
            if isinstance(filepath, Path) or '://' not in filepath:
                filepath = self.root / filepath
            if name != table:
                continue
            load_fn, kwargs = self._load_fn(ext, dask=dask)
            paths = self._resolve_template_vars(filepath)
            if self.use_dask and ext in ('csv', 'json', 'parquet', 'parq') and dask:
                try:
                    df = load_fn(paths, **kwargs)
                except Exception as e:
                    if dask:
                        return self._load_table(table, dask=False)
                    raise e
            else:
                try:
                    dfs = [load_fn(path, **kwargs) for path in paths]
                except Exception as e:
                    if dask:
                        return self._load_table(table, dask=False)
                    raise e
                if len(dfs) <= 1:
                    df = dfs[0] if dfs else None
                elif self.use_dask and hasattr(dfs[0], 'compute'):
                    import dask.dataframe as dd
                    df = dd.concat(dfs)
                else:
                    df = pd.concat(dfs)
            if hasattr(df, 'persist'):
                df = df.persist()
        if df is None:
            tables = list(self._named_files)
            raise ValueError(f"Table '{table}' not found. Available tables include: {tables}.")
        return df

    @cached
    def get(self, table: str, **query) -> DataFrame:
        dask = query.pop('__dask', self.dask)
        df = self._load_table(table)
        df = FilterTransform.apply_to(df, conditions=list(query.items()))
        return df if dask or not hasattr(df, 'compute') else df.compute()


class BaseSQLSource(Source):
    """
    The BaseSQLSource implements the additional API required by
    a SQL based data source.
    """

    dialect = 'any'

    excluded_tables = param.List(default=[], doc="""
        List of table names that should be excluded from the results. Supports:
        - Fully qualified name: 'DATABASE.SCHEMA.TABLE'
        - Schema qualified name: 'SCHEMA.TABLE'
        - Table name only: 'TABLE'
        - Wildcards: 'SCHEMA.*'""")

    load_schema = param.Boolean(default=True, doc="Whether to load the schema")

    # Declare this source supports SQL transforms
    _supports_sql = True

    def __init__(self, **params):
        super().__init__(**params)
        self._exclude_tables_regex = None

    def _is_table_excluded(self, table_slug):
        """
        Check if a table should be excluded based on patterns in self.excluded_tables.
        Case-insensitive matching.
        """
        if not self.excluded_tables:
            return False

        table_slug_lower = table_slug.lower()

        for pattern in self.excluded_tables:
            if not pattern:  # Skip empty patterns
                continue

            pattern_lower = pattern.lower()

            # Check for exact match with full name
            if fnmatch.fnmatch(table_slug_lower, pattern_lower):
                return True

            # Handle cases where we're matching just the table name or schema.table
            parts = table_slug.split('.')
            for i in range(1, len(parts) + 1):
                suffix = '.'.join(parts[-i:])
                if fnmatch.fnmatch(suffix.lower(), pattern_lower):
                    return True

        return False

    def _apply_transforms(self, source: Source, sql_transforms: list[SQLTransform]) -> Source:
        if not sql_transforms:
            return source
        sql_expr = source._sql_expr
        for sql_transform in sql_transforms:
            sql_expr = sql_transform.apply(sql_expr)
        return type(source)(**dict(source._init_args, sql_expr=sql_expr))

    def normalize_table(self, table: str) -> str:
        """
        Allows implementing table name normalization to allow fuzze matching
        of the table name for minor variations such as quoting differences.
        """
        return table

    def get_sql_expr(self, table: str | dict):
        """
        Returns the SQL expression corresponding to a particular table.
        """
        if isinstance(self.tables, dict):
            try:
                table = self.tables[self.normalize_table(table)]
            except KeyError:
                raise KeyError(f"Table {table} not found in {self.tables.keys()}")
        else:
            table = self.normalize_table(table)

        sql_expr = SQLSelectFrom(sql_expr=self.sql_expr).apply(table)
        return sql_expr

    def create_sql_expr_source(self, tables: dict[str, str], **kwargs):
        """
        Creates a new SQL Source given a set of table names and
        corresponding SQL expressions.
        """
        raise NotImplementedError

    def execute(self, sql_query: str, *args, **kwargs) -> pd.DataFrame:
        """
        Executes a SQL query and returns the result as a DataFrame.

        Arguments
        ---------
        sql_query : str
            The SQL Query to execute
        *args : list
            Positional arguments to pass to the SQL query
        **kwargs : dict
            Keyword arguments to pass to the SQL query

        Returns
        -------
        pd.DataFrame
            The result as a pandas DataFrame
        """
        raise NotImplementedError

    @cached_schema
    def get_schema(
        self, table: str | None = None, limit: int | None = None, shuffle: bool = False
    ) -> dict[str, dict[str, Any]] | dict[str, Any]:
        if table is None:
            tables = self.get_tables()
        else:
            tables = [table]

        schemas = {}
        sql_transforms = [SQLSample(size=limit or 1, read=self.dialect)] if shuffle else [SQLLimit(limit=limit or 1, read=self.dialect)]
        for entry in tables:
            if not self.load_schema:
                schemas[entry] = {}
                continue
            sql_expr = self.get_sql_expr(entry)
            data_sql_expr = sql_expr
            for sql_transform in sql_transforms:
                data_sql_expr = sql_transform.apply(data_sql_expr)
            data = self.execute(data_sql_expr)
            schemas[entry] = schema = get_dataframe_schema(data)['items']['properties']

            count_expr = SQLCount(read=self.dialect).apply(sql_expr)
            count_expr = ' '.join(count_expr.splitlines())
            count_data = self.execute(count_expr)
            count_col = 'count' if 'count' in count_data else 'COUNT'
            count = int(count_data[count_col].iloc[0])
            if limit:
                # the min/max and enums will be computed on the limited dataset
                schema['__len__'] = count
                continue

            # patch the min/max and enums from the full dataset
            enums, min_maxes = [], []
            for name, col_schema in schema.items():
                if 'enum' in col_schema:
                    enums.append(name)
                elif 'inclusiveMinimum' in col_schema:
                    min_maxes.append(name)
            for col in enums:
                distinct_expr = SQLDistinct(columns=[col], read=self.dialect).apply(sql_expr)
                distinct_expr = ' '.join(distinct_expr.splitlines())
                distinct = self.execute(distinct_expr)
                schema[col]['enum'] = distinct[col].tolist()

            schema['__len__'] = count
            if not min_maxes:
                continue

            minmax_expr = SQLMinMax(columns=min_maxes, read=self.dialect).apply(sql_expr)
            minmax_expr = ' '.join(minmax_expr.splitlines())
            minmax_data = self.execute(minmax_expr)
            for col in min_maxes:
                kind = data[col].dtype.kind
                if kind in 'iu':
                    cast = int
                elif kind == 'f':
                    cast = float
                elif kind == 'M':
                    cast = str
                else:
                    cast = lambda v: v

                # some dialects, like snowflake output column names to UPPERCASE regardless of input case
                min_col = f'{col}_min' if f'{col}_min' in minmax_data else f'{col}_MIN'
                min_data = minmax_data[min_col].iloc[0]
                max_col = f'{col}_max' if f'{col}_max' in minmax_data else f'{col}_MAX'
                max_data = minmax_data[max_col].iloc[0]
                schema[col]['inclusiveMinimum'] = min_data if pd.isna(min_data) else cast(min_data)
                schema[col]['inclusiveMaximum'] = max_data if pd.isna(max_data) else cast(max_data)

        return schemas if table is None else schemas[table]



class JSONSource(FileSource):
    """
    The JSONSource is very similar to the FileSource but loads json files.

    Both local and remote JSON files can be fetched by declaring them
    as a list or dictionaries of `tables`.
    """

    cache_per_query = param.Boolean(default=False, doc="""
        Whether to query the whole dataset or individual queries.""")

    chunk_size = param.Integer(default=0, doc="""
        Number of items to load per chunk if a template variable
        is provided.""")

    tables = param.ClassSelector(class_=(list, dict), doc="""
        List or dictionary of tables to load. If a list is supplied the
        names are computed from the filenames, otherwise the keys are
        the names. The values must filepaths or URLs to the data:

        ```
        {
            'local' : '/home/user/local_file.csv',
            'remote': 'https://test.com/test.csv'
        }
        ```
    """)

    source_type: ClassVar[str] = 'json'

    def _resolve_template_vars(self, template: str) -> list[str]:
        template_vars = self._template_re.findall(template)
        template_values = []
        for m in template_vars:
            values = state.resolve_reference(f'${m[2:-1]}')
            if not isinstance(values, list):
                values = [values]
            template_values.append(values)
        tables = []
        cross_product = list(product(*template_values))
        if self.chunk_size and len(cross_product) > self.chunk_size:
            for i in range(len(cross_product)//self.chunk_size):
                start = i*self.chunk_size
                chunk = cross_product[start: start+self.chunk_size]
                tvalues = zip(*chunk)
                table = template
                for m, tvals in zip(template_vars, tvalues):
                    tvals = ','.join([v for v in set(tvals)])
                    table = table.replace(m, quote(tvals))
                tables.append(table)
        else:
            table = template
            for m, tvals in zip(template_vars, zip(*cross_product)):
                values = ','.join([v for v in set(tvals)])
                table = table.replace(m, quote(values))
            tables.append(table)
        return tables

    def _load_fn(self, ext: str, dask: bool = True) -> DataFrame:
        return super()._load_fn('json', dask=dask)


class WebsiteSource(Source):
    """
    `WebsiteSource` queries whether a website responds with a 400 status code.
    """

    cache_per_query = param.Boolean(default=False, doc="""
        Whether to query the whole dataset or individual queries.""")

    urls = param.List(doc="URLs of the websites to monitor.")

    source_type: ClassVar[str] = 'live'

    @cached_schema
    def get_schema(
        self, table: str | None = None, limit: int | None = None, shuffle: bool = False
    ) -> dict[str, dict[str, Any]] | dict[str, Any]:
        schema = {
            "status": {
                "url": {"type": "string", 'enum': self.urls},
                "live": {"type": "boolean"}
            }
        }
        return schema if table is None else schema[table]

    def get_tables(self) -> list[str]:
        return ['status']

    @cached
    def get(self, table: str, **query) -> pd.DataFrame:
        data = []
        for url in self.urls:
            try:
                r = requests.get(url)
                live = r.status_code == 200
            except Exception:
                live = False
            data.append({"live": live, "url": url})
        df = pd.DataFrame(data)
        return df


class PanelSessionSource(Source):
    """"
    `PanelSessionSource` queries the session_info endpoint of a Panel application.

    Panel applications with --rest-session-info enabled can be queried
    about session statistics. This source makes this data available to
    Lumen for monitoring.
    """

    cache_per_query = param.Boolean(default=False, doc="""
        Whether to query the whole dataset or individual queries.""")

    endpoint = param.String(default="rest/session_info")

    urls = param.List(doc="URL of the websites to monitor.")

    timeout = param.Parameter(default=5)

    source_type: ClassVar[str] = 'session_info'

    @cached_schema
    def get_schema(
        self, table: str | None = None, limit: int | None = None, shuffle: bool = False
    ) -> dict[str, dict[str, Any]] | dict[str, Any]:
        schema = {
            "summary": {
                "url": {"type": "string", "enum": self.urls},
                "total": {"type": "int"},
                "live": {"type": "int"},
                "render_duration": {"type": "float"},
                "session_duration": {"type": "float"}
            },
            "sessions": {
                "url": {"type": "string", "enum": self.urls},
                "id": {"type": "string"},
                "started": {"type": "float"},
                "ended": {"type": "float"},
                "rendered": {"type": "float"},
                "render_duration": {"type": "float"},
                "session_duration": {"type": "float"},
                "user_agent": {"type": "string"}
            }
        }
        return schema if table is None else schema[table]

    def get_tables(self) -> list[str]:
        return ['summary', 'sessions']

    def _get_session_info(self, table: str, url: str) -> list[dict[str, Any]]:
        res = requests.get(
            url + self.endpoint, verify=False, timeout=self.timeout
        )
        data: list[dict[str, Any]] = []
        if res.status_code != 200:
            return data
        r = res.json()
        session_info = r['session_info']
        sessions = session_info['sessions']

        if table == "summary":
            rendered = [s for s in sessions.values()
                        if s['rendered'] is not None]
            ended = [s for s in sessions.values()
                     if s['ended'] is not None]
            row = {
                'url': url,
                'total': session_info['total'],
                'live': session_info['live'],
                'render_duration': np.mean([s['rendered']-s['started']
                                            for s in rendered]),
                'session_duration': np.mean([s['ended']-s['started']
                                             for s in ended])
            }
            data.append(row)
        elif table == "sessions":
            for sid, session in sessions.items():
                row = dict(url=url, id=sid, **session)
                if session["rendered"]:
                    row["render_duration"] = session["rendered"]-session["started"]
                else:
                    row["render_duration"] = float('NaN')
                if session["ended"]:
                    row["session_duration"] = session["ended"]-session["started"]
                else:
                    row["session_duration"] = float('NaN')
                data.append(row)
        return data

    @cached
    def get(self, table: str, **query) -> pd.DataFrame:
        data = []
        with futures.ThreadPoolExecutor(len(self.urls)) as executor:
            tasks = {executor.submit(self._get_session_info, table, url): url
                     for url in self.urls}
            for future in futures.as_completed(tasks):
                url = tasks[future] + self.endpoint
                try:
                    data.extend(future.result())
                except Exception as e:
                    exception = f"{type(e).__name__}({e})"
                    self.param.warning("Failed to fetch session_info from "
                                       f"{url}, errored with {exception}.")
        return pd.DataFrame(data, columns=list(self.get_schema(table)))


class JoinedSource(Source):
    """
    `JoinedSource` performs a join on tables from one or more sources.

    A JoinedSource applies a join on two or more sources returning
    new table(s) with data from all sources. It iterates over the
    `tables` specification and merges the specified tables from the
    declared sources on the supplied index.

    In this way multiple tables from multiple sources can be merged.
    Individual tables from sources that should not be joined may also
    be surfaced by declaring a single source and table in the
    specification.

    As a simple example we may have sources A and B, which contain
    tables 'foo' and 'bar' respectively. We now want to merge these
    tables on column 'a' in Table A with column 'b' in Table B::

        {'new_table': [
          {'source': 'A', 'table': 'foo', 'index': 'a'},
          {'source': 'B', 'table': 'bar', 'index': 'b'}
        ]}

    The joined source will now publish the "new_table" with all
    columns from tables "foo" and "bar" except for the index column
    from table "bar", which was merged with the index column "a" from
    table "foo".
    """

    sources = param.ClassSelector(class_=(list, dict), doc="""
        A dictionary of sources indexed by their assigned name.""")

    tables = param.Dict(default={}, doc="""
        A dictionary with the names of the joined sources as keys
        and a specification of the source, table and index to merge
        on.

        ```
        {"new_table": [
            {'source': <source_name>,
             'table': <table_name>,
             'index': <index_name>
            },
            {'source': <source_name>,
             'table': <table_name>,
             'index': <index_name>
            },
            ...
        ]}
        ```
        """)

    source_type: ClassVar[str] = 'join'

    def get_tables(self) -> list[str]:
        return list(self.tables)

    @cached_schema
    def get_schema(
        self, table: str | None = None, limit: int | None = None, shuffle: bool = False
    ) -> dict[str, dict[str, Any]] | dict[str, Any]:
        schemas: dict[str, dict[str, Any]] = {}
        for name, specs in self.tables.items():
            if table is not None and name != table:
                continue
            schema: dict[str, Any] = {}
            schemas[name] = schema
            for spec in specs:
                source, subtable = spec['source'], spec['table']
                table_schema = self.sources[source].get_schema(subtable)
                if not schema:
                    schema.update(table_schema)
                else:
                    for column, col_schema in table_schema.items():
                        schema[column] = merge_schemas(col_schema, schema.get(column))
        return schemas if table is None else schemas[table]

    @cached
    def get(self, table: str, **query) -> DataFrame:
        df, left_key = None, None
        for spec in self.tables[table]:
            source, subtable = spec['source'], spec['table']
            source_query = dict(query)
            right_key = spec.get('index')
            if df is not None and left_key and right_key not in query:
                source_query[right_key] = list(df[left_key].unique())
            df_merge = self.sources[source].get(subtable, **source_query)
            if df is None:
                df = df_merge
                left_key = spec.get('index')
            else:
                df = pd.merge(df, df_merge, left_on=left_key,
                              right_on=right_key, how='outer')
        return df  # type: ignore

    @property
    def panel(self) -> pn.Column:
        column = pn.Column(sizing_mode='stretch_width')
        for name, source in self.sources.items():
            panel = source.panel
            if not panel:
                continue
            header = pn.pane.Markdown(f'#### {name.title()}', margin=(0, 5))
            column.extend([header, *source.panel])
        return column

    def clear_cache(self):
        super().clear_cache()
        for source in self.sources.values():
            source.clear_cache()



class DerivedSource(Source):
    """
    `DerivedSource` applies filtering and transforms to tables from other sources.

    A DerivedSource references tables on other sources and optionally
    allows applying filters and transforms to the returned data which
    is then made available as a new (derived) table.

    The DerivedSource has two modes:

    **Table Mode**

    When an explicit `tables` specification is provided full control
    over the exact tables to filter and transform is available. This
    is referred to as the 'table' mode.

    In 'table' mode the tables can reference any table on any source
    using the reference syntax and declare filters and transforms to
    apply to that specific table, e.g. a table specification might
    look like this::

        {
          'derived_table': {
            'source': 'original_source',
            'table': 'original_table'
            'filters': [
              ...
            ],
            'transforms': [
              ...
            ]
          }
        }

    **Mirror mode**

    When a `source` is declared all tables on that Source are mirrored
    and filtered and transformed according to the supplied `filters`
    and `transforms`. This is referred to as 'mirror' mode.

    In mirror mode the DerivedSource may reference an existing source
    directly, e.g.::

        {
            'type': 'derived',
            'source': 'original_source',
            'filters': [...],
            'transforms': [...],
        }
    """

    cache_per_query = param.Boolean(default=False, doc="""
        Whether to query the whole dataset or individual queries.""")

    filters = param.List(doc="""
        A list of filters to apply to all tables of this source.""")

    source = param.ClassSelector(class_=Source, doc="""
        A source to mirror the tables on.""")

    tables = param.Dict(default={}, doc="""
        The dictionary of tables and associated filters and transforms.""")

    transforms = param.List(doc="""
        A list of transforms to apply to all tables of this source.""")

    source_type: ClassVar[str] = 'derived'

    @classmethod
    def _validate_filters(cls, *args, **kwargs) -> list[dict[str, Any] | str]:  # type: ignore
        return cls._validate_list_subtypes('filters', Filter, *args, **kwargs)

    def _get_source_table(self, table: str) -> DataFrame:
        if self.tables:
            spec = self.tables.get(table)
            if spec is None:
                raise ValidationError(
                    f"Table '{table}' was not declared on the DerivedSource. "
                    f"Available tables include {list(self.tables)}."
                )
            source, table = spec['source'], spec['table']
            filters = spec.get('filters', []) + self.filters
        else:
            source = self.source
            filters = self.filters
        query = dict({filt.field: filt.value for filt in filters})
        return source.get(table, **query)

    @cached
    def get(self, table: str, **query) -> DataFrame:
        df = self._get_source_table(table)
        if self.tables:
            transforms = self.tables[table].get('transforms', []) + self.transforms
        else:
            transforms = self.transforms
        transforms.append(FilterTransform(conditions=list(query.items())))
        for transform in transforms:
            df = transform.apply(df)
        return df

    get.__doc__ = Source.get.__doc__

    def get_tables(self) -> list[str]:
        return list(self.tables) if self.tables else self.source.get_tables()

    def clear_cache(self):
        super().clear_cache()
        if self.tables:
            for spec in self.tables.values():
                spec['source'].clear_cache()
        else:
            self.source.clear_cache()

__all__ = [name for name, obj in locals().items() if isinstance(obj, type) and issubclass(obj, Source)]



from __future__ import annotations

import asyncio

from pathlib import Path, PosixPath

import panel as pn
import yaml

from instructor.utils import disable_pydantic_error_url

from ..config import SOURCE_TABLE_SEPARATOR  # NOQA: F401


class LlmSetupError(Exception):
    """
    Raised when an error occurs during the setup of the LLM.
    """


class RetriesExceededError(Exception):
    """
    Raised when the maximum number of retries is exceeded.
    """


THIS_DIR = Path(__file__).parent
PROMPTS_DIR = THIS_DIR / "prompts"

GETTING_STARTED_SUGGESTIONS = [
    "What datasets are available?",
    "What's interesting to analyze?",
    "Can you visualize the data?",
]

DEMO_MESSAGES = [
    "What datasets are available?",
    "Show me the the first dataset and its columns.",
    "What are some interesting analyses I can do?",
    "Perform an analysis using the first suggestion.",
    "Show me a plot of these results.",
]

DEFAULT_EMBEDDINGS_PATH = Path("embeddings")

UNRECOVERABLE_ERRORS = (
    ImportError,
    LlmSetupError,
    RecursionError,
    asyncio.CancelledError,
)

PROVIDED_SOURCE_NAME = "ProvidedSource00000"

# Vega-lite zoomable map configuration
VEGA_ZOOMABLE_MAP_ITEMS = {
    "params": [
        {"name": "tx", "update": "width / 2"},
        {"name": "ty", "update": "height / 2"},
        {
            "name": "scale",
            "value": 300,
            "on": [{"events": {"type": "wheel", "consume": True}, "update": "clamp(scale * pow(1.003, -event.deltaY * pow(16, event.deltaMode)), 150, 50000)"}],
        },
        {"name": "angles", "value": [0, 0], "on": [{"events": "pointerdown", "update": "[rotateX, centerY]"}]},
        {"name": "cloned", "value": None, "on": [{"events": "pointerdown", "update": "copy('projection')"}]},
        {"name": "start", "value": None, "on": [{"events": "pointerdown", "update": "invert(cloned, xy())"}]},
        {"name": "drag", "value": None, "on": [{"events": "[pointerdown, window:pointerup] > window:pointermove", "update": "invert(cloned, xy())"}]},
        {"name": "delta", "value": None, "on": [{"events": {"signal": "drag"}, "update": "[drag[0] - start[0], start[1] - drag[1]]"}]},
        {"name": "rotateX", "value": -240, "on": [{"events": {"signal": "delta"}, "update": "angles[0] + delta[0]"}]},
        {"name": "centerY", "value": 40, "on": [{"events": {"signal": "delta"}, "update": "clamp(angles[1] + delta[1], -60, 60)"}]},
    ],
    "projection": {
        "scale": {"signal": "scale"},
        "rotate": [{"signal": "rotateX"}, 0, 0],
        "center": [0, {"signal": "centerY"}],
        "translate": [{"signal": "tx"}, {"signal": "ty"}]
    }
}

# Map outline layer configurations
VEGA_MAP_LAYER = {
    "world": {
        "data": {
            "url": "https://vega.github.io/vega-datasets/data/world-110m.json",
            "format": {"type": "topojson", "feature": "countries"}
        },
        "mark": {"type": "geoshape", "fill": None, "stroke": "black"}
    },
    "usa": {
        "data": {
            "url": "https://vega.github.io/vega-datasets/data/us-10m.json",
            "format": {"type": "topojson", "feature": "states"}
        },
        "mark": {"type": "geoshape", "fill": None, "stroke": "black"}
    }
}

def str_presenter(dumper, data):
    if "\n" in data:  # Only use literal block for strings containing newlines
        return dumper.represent_scalar("tag:yaml.org,2002:str", data.strip(), style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


def tuple_presenter(dumper, data):
    return dumper.represent_sequence("tag:yaml.org,2002:seq", data)


def path_representer(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', str(data.resolve()))

pn.chat.ChatStep.min_width = 375
pn.chat.ChatStep.collapsed_on_success = False

disable_pydantic_error_url()
yaml.add_representer(str, str_presenter)
yaml.add_representer(tuple, tuple_presenter)
yaml.add_representer(PosixPath, path_representer)
yaml.add_representer(Path, path_representer)



import json
import threading

from typing import Any

import google.auth
import pandas as pd
import param

from google.auth.credentials import Credentials
from google.auth.exceptions import DefaultCredentialsError
from google.cloud import bigquery, exceptions
from google.cloud.bigquery.client import Client
from tqdm import tqdm

from ..transforms.sql import SQLFilter, SQLMinMax
from .base import BaseSQLSource, cached, cached_schema


class BigQuerySource(BaseSQLSource):
    """
    BigQuerySource provides access to a Google BigQuery project.
    """

    datasets = param.List(default=None, doc="List of datasets to include")

    filter_in_sql = param.Boolean(default=True, doc="Whether to apply filters in SQL or in-memory.")

    location = param.String(doc="Location where the project resides.")

    project_id = param.String(doc="The Google Cloud's project ID.")

    tables = param.ClassSelector(class_=(list, dict), doc="""
       A list of tables or a dictionary mapping from table name to a SQL query.""")

    dialect = "bigquery"

    def __init__(self, **params) -> None:
        self._metadata__client: Client | None = None
        self._sql__client: Client | None = None
        self._credentials: Credentials | None = None
        self._cached_tables: list[str] | None = None
        self._metadata_lock = threading.Lock()
        super().__init__(**params)

    def _create_client(
        self,
        project_id: str | None = None,
        location: str | None = None,
    ) -> Client | None:
        client = None
        kwargs = {"credentials": self._credentials}
        if project_id is not None:
            kwargs["project"] = project_id
        if location is not None:
            kwargs["location"] = location
        try:
            client = bigquery.Client(**kwargs)
        except Exception as e:
            msg = "Unable to create a BigQuery client."
            raise exceptions.ClientError(msg) from e

        return client

    @property
    def _sql_client(self):
        if not self._credentials:
            self._authorize()
        self._sql__client = self._create_client(location=self.location)
        return self._sql__client

    @property
    def _metadata_client(self):
        if not self._credentials:
            self._authorize()
        self._metadata__client = self._create_client(project_id=self.project_id, location=self.location)
        return self._metadata__client

    def _filter_dataset(self, dataset: str):
        return False

    def _authorize(self) -> None:
        """Get the default credentials for the current environment.

        To enable application default credentials with the Cloud SDK for the first time, run the
        following manually.

            gcloud init

        To manually enable authentication, run the following.

            gcloud auth application-default login

        This method will attempt to connect to Google Cloud using the SDK after you have run
        `gcloud init`.

        Returns
        -------
        None

        Raises
        ------
        DefaultCredentialsError
            This error is raised if no default credentials can be found. To prevent this in the
            future, you will need to run `gcloud init` at least once.
        """
        try:
            self._credentials, project_id = google.auth.default()  # pyright: ignore [reportAttributeAccessIssue]
        except DefaultCredentialsError:
            msg = (
                "No default credentials can be found. Please run `gcloud init` at least once "
                "to create default credentials to your Google Cloud project."
            )
            raise DefaultCredentialsError(msg)

        if self.project_id == "":
            self.project_id = project_id

    def execute(self, sql_query: str) -> pd.DataFrame:
        return self._sql_client.query_and_wait(sql_query).to_dataframe()

    def get_tables(self) -> list[str]:
        """Get a list of available tables for the project.

        Returns
        -------
        list[str]
            Table names are composed of f"{project_id}.{dataset_id}.{table_id}".
        """
        if self.tables:
            return list(self.tables)
        elif self._cached_tables:
            return self._cached_tables

        table_references = []
        project_id = self._metadata_client.project
        dataset_records = list(self._metadata_client.list_datasets())
        for dataset_record in tqdm(dataset_records):
            dataset_id = dataset_record.dataset_id
            if dataset_id not in self.datasets or self._filter_dataset(dataset_id):
                continue
            table_records = self._metadata_client.list_tables(dataset=dataset_id)
            table_ids = [
                table_record.table_id
                for table_record in table_records
                if table_record.table_id is not None
            ]
            for table_id in table_ids:
                table_references.append(f"{project_id}.{dataset_id}.{table_id}")

        self._cached_tables = table_references
        return table_references

    def get_sql_expr(self, table: str) -> str:
        if isinstance(self.tables, dict):
            if table not in self.tables:
                raise KeyError(f"Table {table} not found in {list(self.tables)}.")
            return self.tables[table]
        return f"SELECT * FROM {table}"

    def create_sql_expr_source(self, tables: dict[str, str], **kwargs):
        params = dict(self.param.values(), **kwargs)
        params.pop("name", None)
        params["tables"] = tables
        source = type(self)(**params)
        return source

    def _get_dataset_metadata(self, datasets: list[str] | None = None) -> dict:
        """Get metadata for all available datasets in the project.

        Parameters
        ----------
        datasets: list[str] | None

        Returns
        -------
        dict
        """
        data = {}
        for record in self._metadata_client.list_datasets():
            dataset_id = record.dataset_id
            if (datasets is not None and dataset_id not in datasets):
                continue
            project_id = record.project
            dataset = self._metadata_client.get_dataset(dataset_ref=f"{project_id}.{dataset_id}")
            data[dataset_id] = {
                "friendly_name": dataset.friendly_name,
                "description": dataset.description,
                "created": str(dataset.created),
                "modified": str(dataset.modified),
            }
        return data

    def _get_tables_metadata(self, project_id: str, dataset_id: str, tables: list[str] | None = None) -> dict:
        """Get metadata for all tables associated with the given dataset.

        Parameters
        ----------
        project_id : str
            The BigQuery project ID.
        dataset_id : str
            The dataset ID within the given project.

        Returns
        -------
        dict
        """
        data = {}
        for record in self._metadata_client.list_tables(dataset=dataset_id):
            table_id = record.table_id
            table_reference = f"{project_id}.{dataset_id}.{table_id}"
            if tables and table_reference not in tables:
                continue
            table = self._metadata_client.get_table(table=table_reference)
            table_clustering_fields = json.dumps(table.clustering_fields)
            if table_clustering_fields == "null":
                table_clustering_fields = None
            table_labels = json.dumps(table.labels)
            if table_labels == "null":
                table_labels = None

            columns = {}
            for column in table.schema:
                column_fields = []
                for column_field in column.fields:
                    field_datum = column_field.__dict__["_properties"]
                    column_fields.append(field_datum)
                if column_fields:
                    column_fields = json.dumps(column_fields)
                else:
                    column_fields = None
                columns[column.name] = {
                    "type": column.field_type,
                    "description": column.description,
                    "fields": column_fields,
                }
            data[table_reference] = {
                "columns": columns,
                "friendly_name": table.friendly_name,
                "description": table.description,
                "created": str(table.created),
                "modified": str(table.modified),
                "num_rows": table.num_rows,
                "clustering_fields": table_clustering_fields,
                "labels": table_labels,
            }
        return data

    def _get_table_metadata(self, tables: list[str]) -> dict:
        if self.datasets is None:
            datasets = [record.dataset_id for record in self._metadata_client.list_datasets()]
        else:
            datasets = self.datasets
        metadata = {}
        for dataset in datasets:
            if self.datasets is None or dataset in self.datasets:
                metadata.update(self._get_tables_metadata(self.project_id, dataset, tables))
        return metadata

    def _get_table(self, table: str):
        with self._metadata_lock:
            return self._metadata_client.get_table(table=table)

    def _get_table_schema(
        self,
        table: str,
        limit: int | None = None,
        shuffle: bool = False,
    ) -> dict[str, dict[str, Any]] | dict[str, Any]:
        schema = {}
        bq_table = self._get_table(table)
        enums = []
        min_maxes = []
        for column in bq_table.schema:
            column_name = column.name
            if column_name in schema:
                continue
            column_type = column.field_type

            if column_type == "STRING":
                schema[column_name] = {"type": "string", "enum": []}
                enums.append(column_name)
            elif column_type in ["TIMESTAMP", "DATE"]:
                schema[column_name] = {
                    "type": "str",
                    "format": "datetime",
                    "min": "",
                    "max": "",
                }
                min_maxes.append(column_name)
            elif column_type in ["FLOAT", "INTEGER"]:
                schema[column_name] = {
                    "type": "integer" if column_type == "INTEGER" else "number",
                }
                min_maxes.append(column_name)
            elif column_type == "BOOLEAN":
                schema[column_name] = {"type": column_type.lower()}
            elif column_type == "RECORD":
                print(f"{column_name} is of type {column_type}, which we currently ignore.")
                continue

        sql_expr = self.get_sql_expr(table)
        for col in enums:
            distinct_expr = f"SELECT DISTINCT {col} FROM {sql_expr} LIMIT 1000"
            distinct = self.execute(distinct_expr)
            schema[col]['enum'] = distinct[col].tolist()

        if not min_maxes:
            return schema

        minmax_expr = SQLMinMax(columns=min_maxes, read=self.dialect).apply(sql_expr)
        minmax_expr = ' '.join(minmax_expr.splitlines())
        minmax_data = self.execute(minmax_expr)
        for col in min_maxes:
            kind = minmax_data[col].dtype.kind
            if kind in 'iu':
                cast = int
            elif kind == 'f':
                cast = float
            elif kind == 'M':
                cast = str
            else:
                cast = lambda v: v

            # some dialects, like snowflake output column names to UPPERCASE regardless of input case
            min_col = f'{col}_min' if f'{col}_min' in minmax_data else f'{col}_MIN'
            min_data = minmax_data[min_col].iloc[0]
            max_col = f'{col}_max' if f'{col}_max' in minmax_data else f'{col}_MAX'
            max_data = minmax_data[max_col].iloc[0]
            schema[col]['inclusiveMinimum'] = min_data if pd.isna(min_data) else cast(min_data)
            schema[col]['inclusiveMaximum'] = max_data if pd.isna(max_data) else cast(max_data)
        return schema

    @cached_schema
    def get_schema(
        self, table: str | None = None, limit: int | None = None, shuffle: bool = False
    ) -> dict[str, dict[str, Any]] | dict[str, Any]:
        """Determine the schema of the given `table`.

        This method overrides the inherited `get_schema` from the base class `Source`. The reason
        why we override

        Parameters
        ----------
        table : str | None
            The name of the table. Must be in the form: f"{project_id}.{dataset_id}.{table_id}"
            or reference a table in the tables dictionary.
        limit : int | None
            The maximum number of rows to sample from the table.
        shuffle : bool
            Whether to shuffle the rows of the table.

        Returns
        -------
        dict[str, dict[str, Any]] | dict[str, Any]
        """
        tables = [table] if table else self.get_tables()
        schemas = {}
        for t in tables:
            if isinstance(self.tables, dict) or limit:
                # BigQuery sampling is extremely slow, so we don't shuffle
                schemas[t] = super().get_schema(t, limit, False)
            else:
                schemas[t] = self._get_table_schema(t)
        return schemas[table] if table else schemas

    @cached
    def get(self, table, **query):
        query.pop("__dask", None)
        sql_expr = self.get_sql_expr(table)
        sql_transforms = query.pop("sql_transforms", [])
        conditions = list(query.items())
        filter_in_sql = bool(self.filter_in_sql)
        if filter_in_sql:
            sql_transforms = [SQLFilter(conditions=conditions)] + sql_transforms
        for st in sql_transforms:
            sql_expr = st.apply(sql_expr)
        return self.execute(sql_expr)



import os

import param


class DbtslMixin(param.Parameterized):

    auth_token = param.String(default=None, doc="""
        The auth token for the dbt semantic layer client;
        if not provided will fetched from DBT_AUTH_TOKEN env var.""")

    environment_id = param.Integer(default=None, doc="""
        The environment ID for the dbt semantic layer client.""")

    host = param.String(default="semantic-layer.cloud.getdbt.com", doc="""
        The host for the dbt semantic layer client.""")

    def __init__(self, environment_id: int, **params):

        super().__init__(environment_id=environment_id, **params)

    def _get_dbtsl_client(self):
        from dbtsl.asyncio import AsyncSemanticLayerClient
        dbtsl_client = AsyncSemanticLayerClient(
            environment_id=self.environment_id,
            auth_token=self.auth_token or os.getenv("DBT_AUTH_TOKEN"),
            host=self.host,
        )
        return dbtsl_client



import pandas as pd
import pytest

from hvplot.tests.util import makeMixedDataFrame


@pytest.fixture
def source_tables():
    string = pd.get_option('mode.string_storage')
    pd.set_option('mode.string_storage', 'pyarrow')
    df_test = makeMixedDataFrame()
    df_test_sql = makeMixedDataFrame()
    df_test_sql_none = makeMixedDataFrame()
    df_test_sql_none['C'] = ['foo1', None, 'foo3', None, 'foo5']
    tables = {
        'test': df_test,
        'test_sql': df_test_sql,
        'test_sql_with_none': df_test_sql_none,
    }
    yield tables
    pd.set_option('mode.string_storage', string)



import param
import pytest

try:
    import lumen.ai
except ModuleNotFoundError:
    pytest.skip("lumen.ai could not be imported, skipping tests.", allow_module_level=True)

from pydantic import BaseModel, Field

from lumen.ai.translate import param_to_pydantic, pydantic_to_param_instance

try:
    import lumen.ai  # noqa
except ModuleNotFoundError:
    pytest.skip("lumen.ai could not be imported, skipping tests.", allow_module_level=True)

from lumen.ai.translate import doc_descriptions, function_to_model


def add(a: int, b: int) -> int:
    """
    Adds two integers

    Parameters
    ----------
    a : int
        Integer A
    b : int
        Integer B

    Returns
    -------
    int
        Result of addition
    """
    return a + b


def add_google(a: int, b: int) -> int:
    """
    Adds two integers

    Args:
        a (int): Integer A
        b (int): Integer B

    Returns:
        int: Result of addition
    """
    return a + b


def add_sphinx(a: int, b: int) -> int:
    """
    Adds two integers

    :param a: Integer A
    :type a: int
    :param b: Integer B
    :type b: int
    :return: The sum of the two integers.
    :rtype: int
    """
    return a + b


def test_doc_descriptions_numpy():
    main, descriptions = doc_descriptions(add)
    assert main == "Adds two integers"
    assert descriptions["a"] == "Integer A"
    assert descriptions["b"] == "Integer B"


def test_doc_descriptions_google():
    main, descriptions = doc_descriptions(add_google)
    assert main == "Adds two integers"
    assert descriptions["a"] == "Integer A"
    assert descriptions["b"] == "Integer B"


def test_doc_descriptions_sphinx():
    main, descriptions = doc_descriptions(add_sphinx)
    assert main == "Adds two integers"
    assert descriptions["a"] == "Integer A"
    assert descriptions["b"] == "Integer B"


def test_function_to_model_args():
    model = function_to_model(add)

    fields = model.model_fields
    assert len(fields) == 2
    assert "a" in fields
    assert fields["a"].annotation is int
    assert fields["a"].description == "Integer A"
    assert "b" in fields
    assert fields["b"].annotation is int
    assert fields["b"].description == "Integer B"


def test_param_to_pydantic_conversion():
    """
    Test that param.Parameterized classes are correctly converted to Pydantic models.
    This tests our use of Field() instead of FieldInfo in the translate module.
    """

    # Define direct Pydantic models for comparison
    class PydanticPet(BaseModel):
        name: str = Field(..., description="Name of the pet")
        age: int = Field(..., ge=0, le=20, description="Age of the pet in years")

    class PydanticDog(PydanticPet):
        bark_volume: int = Field(default=5, ge=0, le=10)

    class PydanticCat(PydanticPet):
        meow_pitch: float = Field(default=0.5, gt=0, lt=1)

    class PydanticOwner(BaseModel):
        pet: PydanticDog | PydanticCat

    # Define param.Parameterized classes
    class Pet(param.Parameterized):
        name = param.String(doc="Name of the pet")
        age = param.Integer(default=1, bounds=(0, 20), doc="Age of the pet in years")

    class Dog(Pet):
        bark_volume = param.Integer(default=5, bounds=(0, 10))

    class Cat(Pet):
        meow_pitch = param.Number(default=0.5, bounds=(0, 1))

    class Owner(param.Parameterized):
        pet = param.ClassSelector(class_=(Dog, Cat))

    # Convert param classes to Pydantic models
    created_models = param_to_pydantic(Owner, excluded=["name"])
    ParamOwner = created_models["Owner"]
    ParamDog = created_models["Dog"]
    ParamCat = created_models["Cat"]

    # Get schemas
    param_owner_schema = ParamOwner.model_json_schema()

    # Test pet field in Owner
    assert "pet" in param_owner_schema["properties"]

    # Test field metadata conversion in Dog
    param_dog_schema = ParamDog.model_json_schema()
    assert "bark_volume" in param_dog_schema["properties"]
    bark_volume_prop = param_dog_schema["properties"]["bark_volume"]
    assert bark_volume_prop["default"] == 5
    assert "minimum" in bark_volume_prop
    assert "maximum" in bark_volume_prop
    assert bark_volume_prop["minimum"] == 0
    assert bark_volume_prop["maximum"] == 10

    # Test field metadata conversion in Cat
    param_cat_schema = ParamCat.model_json_schema()
    assert "meow_pitch" in param_cat_schema["properties"]
    meow_pitch_prop = param_cat_schema["properties"]["meow_pitch"]
    assert meow_pitch_prop["default"] == 0.5
    assert "minimum" in meow_pitch_prop
    assert "maximum" in meow_pitch_prop
    assert meow_pitch_prop["minimum"] == 0
    assert meow_pitch_prop["maximum"] == 1

    # Test description preservation
    param_pet_schema = param_dog_schema  # Dog inherits from Pet
    assert "age" in param_pet_schema["properties"]
    age_prop = param_pet_schema["properties"]["age"]
    assert "description" in age_prop
    assert age_prop["description"] == "Age of the pet in years"


def test_field_bounds_conversion():
    """
    Test that parameter bounds are correctly converted to Field constraints.
    """

    class NumberModel(param.Parameterized):
        value = param.Number(default=5, bounds=(0, 10))

    created_models = param_to_pydantic(NumberModel)
    PydanticNumberModel = created_models["NumberModel"]

    # Create an instance to verify constraints work
    model = PydanticNumberModel(value=5)
    assert model.value == 5

    # Check schema
    schema = PydanticNumberModel.model_json_schema()
    assert "value" in schema["properties"]
    value_prop = schema["properties"]["value"]
    assert value_prop["default"] == 5
    assert value_prop["minimum"] == 0
    assert value_prop["maximum"] == 10

    # Verify Field constraints by trying invalid values
    with pytest.raises(Exception):
        PydanticNumberModel(value=-1)  # Below minimum

    with pytest.raises(Exception):
        PydanticNumberModel(value=11)  # Above maximum


def test_alias_handling():
    """
    Test that reserved field names are correctly aliased.
    """

    class ReservedNamesModel(param.Parameterized):
        schema = param.String(default="my_schema")
        copy = param.Boolean(default=True)

    created_models = param_to_pydantic(ReservedNamesModel)
    PydanticReservedModel = created_models["ReservedNamesModel"]

    # First, verify the default values
    model_with_defaults = PydanticReservedModel()
    assert model_with_defaults.schema_ == "my_schema"
    assert model_with_defaults.copy_ == True

    # Create a model with custom values using the field aliases
    # The field name is schema_ but the alias is schema
    model = PydanticReservedModel()
    model.schema_ = "test_schema"
    model.copy_ = False

    # Verify the values were set correctly
    assert model.schema_ == "test_schema"  # Field renamed to schema_
    assert model.copy_ == False  # Field renamed to copy_

    # Check schema to verify aliases
    schema = PydanticReservedModel.model_json_schema()
    assert "schema_" in schema["properties"]
    assert "copy_" in schema["properties"]


def test_various_param_types_conversion():
    """
    Test that various param types are correctly converted to Field objects.
    """

    class AllParamTypes(param.Parameterized):
        # Basic types
        string_param = param.String(default="hello", doc="A string parameter")
        int_param = param.Integer(default=42, bounds=(0, 100), doc="An integer parameter")
        float_param = param.Number(default=3.14, bounds=(0.0, 10.0), doc="A float parameter")
        bool_param = param.Boolean(default=True, doc="A boolean parameter")

        # Collection types
        list_param = param.List(default=[1, 2, 3], doc="A list parameter")
        dict_param = param.Dict(default={"a": 1, "b": 2}, doc="A dictionary parameter")
        tuple_param = param.Tuple(default=(1, "a"), doc="A tuple parameter")

        # Selection types
        selector_param = param.Selector(default="option1", objects=["option1", "option2", "option3"], doc="A selector parameter")

        # Date and time
        date_param = param.Date(default=None, doc="A date parameter")

        # Allow None
        nullable_param = param.String(default=None, allow_None=True, doc="A nullable string parameter")

        # Color param
        color_param = param.Color(default="#FF0000", doc="A color parameter")

    created_models = param_to_pydantic(AllParamTypes)
    PydanticAllParams = created_models["AllParamTypes"]

    # Create an instance with defaults - providing values for the required fields
    instance = PydanticAllParams(date_param=None, nullable_param=None)

    # Check basic types
    assert instance.string_param == "hello"
    assert instance.int_param == 42
    assert instance.float_param == 3.14
    assert instance.bool_param == True

    # Check collection types
    assert instance.list_param == [1, 2, 3]
    assert instance.dict_param == {"a": 1, "b": 2}
    assert instance.tuple_param == (1, "a")

    # Check selector
    assert instance.selector_param == "option1"

    # Check nullable param
    assert instance.nullable_param is None

    # Check schema for bounds and descriptions
    schema = PydanticAllParams.model_json_schema()

    # Check string param
    string_prop = schema["properties"]["string_param"]
    assert string_prop["type"] == "string"
    assert string_prop["default"] == "hello"
    assert string_prop["description"] == "A string parameter"

    # Check int param with bounds
    int_prop = schema["properties"]["int_param"]
    assert int_prop["type"] == "integer"
    assert int_prop["default"] == 42
    assert int_prop["minimum"] == 0
    assert int_prop["maximum"] == 100
    assert int_prop["description"] == "An integer parameter"

    # Check selector param becomes enum
    selector_prop = schema["properties"]["selector_param"]
    assert selector_prop["default"] == "option1"
    assert "enum" in selector_prop
    assert selector_prop["enum"] == ["option1", "option2", "option3"]

    # Check nullable param allows null
    nullable_prop = schema["properties"]["nullable_param"]
    # In Pydantic v2, nullable types might be represented as anyOf/oneOf
    # or with a nullable field instead of type: ["string", "null"]
    if "type" in nullable_prop and isinstance(nullable_prop["type"], list):
        assert "null" in nullable_prop["type"]
    elif "anyOf" in nullable_prop:
        # Check if any schema in anyOf has type null
        null_schema_exists = any(schema.get("type") == "null" for schema in nullable_prop["anyOf"])
        assert null_schema_exists, "No null type found in anyOf schemas"
    elif "nullable" in nullable_prop:
        assert nullable_prop["nullable"] is True


def test_private_attributes():
    """
    Test that attributes with leading underscores are correctly converted to
    Pydantic PrivateAttr objects.
    """

    class ModelWithPrivateAttrs(param.Parameterized):
        _private = param.String(default="private_value")
        public = param.String(default="public_value")

    created_models = param_to_pydantic(ModelWithPrivateAttrs)
    PydanticModel = created_models["ModelWithPrivateAttrs"]

    # Create an instance
    model = PydanticModel()

    # Check that private attributes are directly accessible as attributes
    assert hasattr(model, "_private")
    assert model._private.default == "private_value"
    assert model.public == "public_value"

    # Verify the model schema does NOT include the private attribute
    schema = PydanticModel.model_json_schema()
    assert "_private" not in schema["properties"]
    assert "public" in schema["properties"]


# Round Trip Tests
def test_basic_round_trip():
    """
    Test a basic round trip conversion: param -> pydantic -> param
    """

    # Define a simple Parameterized class
    class SimpleParamClass(param.Parameterized):
        name = param.String(default="test", doc="A test parameter")
        count = param.Integer(default=42, bounds=(0, 100))
        status = param.Boolean(default=True)

    # Create a param instance to start with
    original_param = SimpleParamClass(name="original", count=50, status=False)

    # Convert to pydantic
    created_models = param_to_pydantic(SimpleParamClass)
    PydanticModel = created_models["SimpleParamClass"]

    # Convert back to param
    round_trip_param = pydantic_to_param_instance(PydanticModel(
        name=original_param.name,
        count=original_param.count,
        status=original_param.status
    ))

    # Verify values are preserved
    assert isinstance(round_trip_param, SimpleParamClass)
    assert round_trip_param.name == original_param.name
    assert round_trip_param.count == original_param.count
    assert round_trip_param.status == original_param.status

    # Verify param behaviors are preserved (e.g., bounds checking)
    with pytest.raises(ValueError):
        round_trip_param.count = 101  # Above maximum

    with pytest.raises(ValueError):
        round_trip_param.count = -1  # Below minimum


def test_complex_round_trip():
    """
    Test round trip with complex hierarchy and all parameter types.
    """

    # Define a hierarchy of Parameterized classes
    class Animal(param.Parameterized):
        name = param.String(default="unnamed", doc="Animal name")
        age = param.Integer(default=1, bounds=(0, 30), doc="Age in years")

    class Dog(Animal):
        breed = param.ObjectSelector(default="mixed", objects=["mixed", "labrador", "shepherd"], doc="Dog breed")
        size = param.Selector(default="medium", objects=["small", "medium", "large"], doc="Size category")

    class Person(param.Parameterized):
        name = param.String(default="John Doe", doc="Person name")
        pets = param.List(default=[], item_type=Animal, doc="List of pets")
        favorite_colors = param.List(default=["blue", "green"], doc="Favorite colors")
        preferences = param.Dict(default={"likes_coffee": True, "likes_tea": False}, doc="Preferences")
        birthday = param.Date(default=None, allow_None=True, doc="Birthday")
        profile_color = param.Color(default="#1A85FF", doc="Profile color")

    # Create original param instances
    dog1 = Dog(name="Rex", age=3, breed="labrador", size="large")
    dog2 = Dog(name="Bella", age=2, breed="shepherd", size="medium")
    person = Person(
        name="Alice Smith",
        pets=[dog1, dog2],
        favorite_colors=["purple", "orange"],
        preferences={"likes_coffee": False, "likes_hiking": True},
        birthday=None,
        profile_color="#ff5733",
    )

    # Convert class to pydantic model
    created_models = param_to_pydantic(Person)
    PydanticPerson = created_models["Person"]
    PydanticDog = created_models["Dog"]

    # Convert original dogs to pydantic
    pydantic_dog1 = PydanticDog(name="Rex", age=3, breed="labrador", size="large")

    pydantic_dog2 = PydanticDog(name="Bella", age=2, breed="shepherd", size="medium")

    # Create pydantic person instance manually
    pydantic_person = PydanticPerson(
        name="Alice Smith",
        pets=[pydantic_dog1, pydantic_dog2],
        favorite_colors=["purple", "orange"],
        preferences={"likes_coffee": False, "likes_hiking": True},
        birthday=None,
        profile_color="#ff5733",
    )

    # Convert back to param
    round_trip_person = pydantic_to_param_instance(pydantic_person)

    # Verify all attributes are preserved
    assert isinstance(round_trip_person, Person)
    assert round_trip_person.name == person.name
    assert len(round_trip_person.pets) == 2
    assert round_trip_person.profile_color == "#ff5733"

    # Verify nested objects
    pet1 = round_trip_person.pets[0]
    assert isinstance(pet1, Dog)
    assert pet1.name == "Rex"
    assert pet1.breed == "labrador"
    assert pet1.size == "large"

    # Verify param behavior is preserved
    with pytest.raises(ValueError):
        pet1.age = 31  # Above maximum

    # Verify selector constraint is preserved
    with pytest.raises(ValueError):
        pet1.size = "extra-large"  # Not in allowed values


def test_round_trip_modifications():
    """
    Test round trip with modifications in the pydantic step.
    """

    # Define a param class
    class Config(param.Parameterized):
        name = param.String(default="default-config")
        timeout = param.Integer(default=30, bounds=(1, 300))
        retry_count = param.Integer(default=3, bounds=(0, 10))
        debug_mode = param.Boolean(default=False)
        log_levels = param.ListSelector(default=["error"], objects=["error", "warn", "info", "debug"])

    # Convert to pydantic
    created_models = param_to_pydantic(Config)
    PydanticConfig = created_models["Config"]

    # Create pydantic instance with different values
    pydantic_config = PydanticConfig(name="modified-config", timeout=60, retry_count=5, debug_mode=True, log_levels=["error", "warn", "debug"])

    # Now convert back to param
    round_trip_config = pydantic_to_param_instance(pydantic_config)

    # Verify modified values are preserved
    assert round_trip_config.name == "modified-config"
    assert round_trip_config.timeout == 60
    assert round_trip_config.retry_count == 5
    assert round_trip_config.debug_mode is True
    assert round_trip_config.log_levels == ["error", "warn", "debug"]

    # Verify param behavior is still intact
    with pytest.raises(ValueError):
        round_trip_config.retry_count = 11  # Above maximum

    # Make sure we can modify valid values
    round_trip_config.retry_count = 7
    assert round_trip_config.retry_count == 7


def test_round_trip_with_default_values():
    """
    Test round trip conversion using default values.
    """

    # Define param class with defaults
    class Settings(param.Parameterized):
        name = param.String(default="default-settings")
        max_connections = param.Integer(default=100, bounds=(10, 1000))
        enable_cache = param.Boolean(default=True)
        buffer_size = param.Integer(default=4096)

    # Convert to pydantic model
    created_models = param_to_pydantic(Settings)
    PydanticSettings = created_models["Settings"]

    # Create a pydantic instance with default values
    pydantic_settings = PydanticSettings()

    # Convert back to param
    round_trip_settings = pydantic_to_param_instance(pydantic_settings)

    # Verify default values are preserved
    assert isinstance(round_trip_settings, Settings)
    assert round_trip_settings.name == "default-settings"
    assert round_trip_settings.max_connections == 100
    assert round_trip_settings.enable_cache is True
    assert round_trip_settings.buffer_size == 4096

    # Verify param behavior is preserved
    with pytest.raises(ValueError):
        round_trip_settings.max_connections = 5  # Below minimum



from __future__ import annotations

import asyncio
import logging
import traceback

from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING, Any

import param

from panel.chat import ChatFeed, ChatInterface, ChatMessage
from panel.chat.feed import PLACEHOLDER_SVG
from panel.config import config, panel_extension
from panel.io.document import hold
from panel.io.resources import CSS_URLS
from panel.io.state import state
from panel.layout import (
    Column, HSpacer, Row, Tabs,
)
from panel.pane import SVG, Markdown
from panel.param import ParamMethod
from panel.template import FastListTemplate
from panel.theme import Material
from panel.util import edit_readonly
from panel.viewable import Viewer
from panel.widgets import Button, FileDownload, MultiChoice

from ..config import SOURCE_TABLE_SEPARATOR
from ..pipeline import Pipeline
from ..sources import Source
from ..sources.duckdb import DuckDBSource
from ..transforms.sql import SQLLimit
from ..util import detect_file_encoding, log
from .agents import (
    AnalysisAgent, AnalystAgent, ChatAgent, DocumentListAgent, SourceAgent,
    SQLAgent, TableListAgent, VegaLiteAgent,
)
from .components import SplitJS, StatusBadge
from .config import PROVIDED_SOURCE_NAME
from .controls import SourceControls
from .coordinator import Coordinator, Planner
from .export import (
    export_notebook, make_md_cell, make_preamble, render_cells, write_notebook,
)
from .llm import Llm, OpenAI
from .memory import _Memory, memory
from .tools import TableLookup
from .vector_store import VectorStore

if TYPE_CHECKING:
    from .views import LumenOutput

DataT = str | Path | Source | Pipeline

EXPLORATIONS_INTRO = """
🧪 **Explorations**

- Explorations are analyses applied to one or more datasets.
- An exploration consists of interactive tables and visualizations.
- New explorations are launched in a new tab for each SQL query result.
- Each exploration maintains its own context, so you can branch off an existing result by navigating to that tab.
"""

class UI(Viewer):
    """
    UI provides a baseclass and high-level entrypoint to start chatting with your data.
    """

    analyses = param.List(default=[], doc="""
        List of custom analyses. If provided the AnalysesAgent will be added."""
    )

    coordinator = param.ClassSelector(
        class_=Coordinator, default=Planner, is_instance=False, doc="""
        The Coordinator class that will be responsible for coordinating the Agents."""
    )

    coordinator_params = param.Dict(
        default={}, doc="""Keyword arguments to pass to the Coordinator."""
    )

    agents = param.List(default=[], doc="""
        List of additional Agents to add beyond the default_agents."""
    )

    default_agents = param.List(default=[
        TableListAgent, ChatAgent, DocumentListAgent, AnalystAgent, SourceAgent, SQLAgent, VegaLiteAgent
    ], doc="""List of default agents which will always be added.""")

    export_functions = param.Dict(default={}, doc="""
       Dictionary mapping from name of exporter to export function.""")

    llm = param.ClassSelector(class_=Llm, default=OpenAI(), doc="""
        The LLM provider to be used by default""")

    log_level = param.ObjectSelector(default='INFO', objects=['DEBUG', 'INFO', 'WARNING', 'ERROR'], doc="""
        The log level to use.""")

    logs_db_path = param.String(default=None, doc="""
        The path to the log file that will store the messages exchanged with the LLM.""")

    interface = param.ClassSelector(class_=ChatFeed, doc="""
        The interface for the Coordinator to interact with.""")

    notebook_preamble = param.String(default='', doc="""
        Preamble to add to exported notebook(s).""")

    template = param.Selector(
        default=config.param.template.names['fast'],
        objects=config.param.template.names, doc="""
        Panel template to serve the application in."""
    )

    title = param.String(default='Lumen UI', doc="Title of the app.")

    tools = param.List(doc="""
       List of Tools that can be invoked by the coordinator.""")

    vector_store = param.ClassSelector(
        class_=VectorStore, default=None, doc="""
        The vector store to use for tools unrelated to documents. If not provided, a new one will be created
        or inferred from the tools provided."""
    )

    document_vector_store = param.ClassSelector(
        class_=VectorStore, default=None, doc="""
        The vector store to use for document-related tools. If not provided, a new one will be created
        or inferred from the tools provided."""
    )

    __abstract = True

    def __init__(
        self,
        data: DataT | list[DataT] | None = None,
        **params
    ):
        params["log_level"] = params.get("log_level", self.param["log_level"].default).upper()
        super().__init__(**params)
        log.setLevel(self.log_level)

        agents = self.agents
        agent_types = {type(agent) for agent in agents}
        for default_agent in self.default_agents:
            if default_agent not in agent_types:
                # only add default agents if they are not already set by the user
                agents.append(default_agent)

        if self.analyses:
            agents.append(AnalysisAgent(analyses=self.analyses))

        self._resolve_data(data)

        if self.interface is None:
            self.interface = ChatInterface(
                load_buffer=5, sizing_mode="stretch_both"
            )
        levels = logging.getLevelNamesMapping()
        if levels.get(self.log_level) < 20:
            self.interface.callback_exception = "verbose"
        self.interface.disabled = True
        self._coordinator = self.coordinator(
            agents=agents,
            interface=self.interface,
            llm=self.llm,
            tools=self.tools,
            logs_db_path=self.logs_db_path,
            within_ui=True,
            vector_store=self.vector_store,
            document_vector_store=self.document_vector_store,
            **self.coordinator_params
        )
        self._notebook_export = FileDownload(
            icon="notebook",
            icon_size="1.5em",
            button_type="primary",
            callback=self._export_notebook,
            filename=f"{self.title.replace(' ', '_')}.ipynb",
            stylesheets=['.bk-btn a { padding: 0 6px; }'],
            styles={'z-index': '1000'}
        )
        self._exports = Row(
            HSpacer(),
            self._notebook_export, *(
                Button(
                    name=label, button_type='primary',
                    on_click=lambda _: e(self.interface),
                    stylesheets=['.bk-btn { padding: 4.5px 6px;']
                )
                for label, e in self.export_functions.items()
            ),
            styles={'position': 'relative', 'right': '20px', 'top': '-1px'},
            sizing_mode='stretch_width'
        )
        self._main = Column(self._exports, self._coordinator, sizing_mode='stretch_both')
        self._vector_store_status_badge = StatusBadge(name="Tables Vector Store Pending")
        if state.curdoc and state.curdoc.session_context:
            state.on_session_destroyed(self._destroy)
        state.onload(self._verify_llm)

    async def _verify_llm(self):
        try:
            await self.llm.initialize(log_level=self.log_level)
            self.interface.disabled = False
        except Exception:
            traceback.print_exc()

        for tool in self._coordinator._tools["main"]:
            if isinstance(tool, TableLookup):
                table_lookup = tool
                break

        if not table_lookup:
            self._vector_store_status_badge.param.update(
                status="success", name='Tables Vector Store Ready')
            return

        self._vector_store_status_badge.status = "running"
        while table_lookup._ready is False:
            await asyncio.sleep(0.5)

        if table_lookup._ready:
            self._vector_store_status_badge.param.update(
                status="success", name='Tables Vector Store Ready')
        elif table_lookup._ready is None:
            self._vector_store_status_badge.param.update(
                status="danger", name='Tables Vector Store Error')

    def _destroy(self, session_context):
        """
        Cleanup on session destroy
        """

    def _export_notebook(self):
        nb = export_notebook(self.interface.objects, preamble=self.notebook_preamble)
        return StringIO(nb)

    def _resolve_data(self, data: DataT | list[DataT] | None):
        if data is None:
            return
        elif not isinstance(data, list):
            data = [data]
        sources = []
        mirrors, tables = {}, {}
        remote = False
        for src in data:
            if isinstance(src, Source):
                sources.append(src)
            elif isinstance(src, Pipeline):
                mirrors[src.name] = src
            elif isinstance(src, (str, Path)):
                src = str(src)
                if src.startswith('http'):
                    remote = True
                if src.endswith(('.parq', '.parquet')):
                    table = f"read_parquet('{src}')"
                elif src.endswith(".csv"):
                    encoding = detect_file_encoding(file_obj=src)
                    table = f"read_csv('{src}', encoding='{encoding}')"
                elif src.endswith(".json"):
                    table = f"read_json_auto('{src}')"
                else:
                    raise ValueError(
                        f"Could not determine how to load {src} file."
                    )
                tables[src] = table
        if tables or mirrors:
            initializers = ["INSTALL httpfs;", "LOAD httpfs;"] if remote else []
            source = DuckDBSource(
                tables=tables, mirrors=mirrors,
                uri=':memory:', initializers=initializers,
                name=PROVIDED_SOURCE_NAME
            )
            sources.append(source)
        memory['sources'] = sources
        if sources:
            memory['source'] = sources[0]

    def show(self, **kwargs):
        return self._create_view(server=True).show(**kwargs)

    def _create_view(self, server: bool | None = None):
        config.css_files.append(CSS_URLS['font-awesome'])
        if (state.curdoc and state.curdoc.session_context) or server is True:
            panel_extension(
                *{ext for agent in self._coordinator.agents for ext in agent._extensions}, template=self.template
            )
            config.template = self.template
            template = state.template
            if isinstance(template, FastListTemplate):
                template.main_layout = None
            template.favicon = "https://lumen.holoviz.org/_static/favicon.ico"
            template.title = self.title
            template.config.css_files.append(
                'https://fonts.googleapis.com/css2?family=Nunito:wght@700'
            )
            template.config.raw_css = [
                "#header a { font-family: 'Nunito', sans-serif; font-size: 2em; font-weight: bold;}"
                "#main { padding: 0 } .main .card-margin.stretch_both { margin: 0; height: 100%; }"
            ]
            template.header.append(
                Row(
                    self.llm.status(),
                    self._vector_store_status_badge,
                    sizing_mode="stretch_width",
                    align="center",
                )
            )
            template.main.append(self._main)
            return template
        return super()._create_view()

    def servable(self, title: str | None = None, **kwargs):
        if (state.curdoc and state.curdoc.session_context):
            self._create_view()
            return self
        return self._create_view().servable(title, **kwargs)

    def __panel__(self):
        return self._main

    def _repr_mimebundle_(self, include=None, exclude=None):
        panel_extension(
            *{ext for exts in self._coordinator.agents for ext in exts}, design='material', notifications=True
        )
        return self._create_view()._repr_mimebundle_(include, exclude)


class ChatUI(UI):
    """
    ChatUI provides a high-level entrypoint to start chatting with your data
    in a chat based UI.

    This high-level wrapper allows providing the data sources you will
    be chatting with and then configures the assistant and agents.

    :Example:

    .. code-block:: python

        import lumen.ai as lmai

        lmai.ChatUI(data='~/data.csv').servable()
    """

    title = param.String(default='Lumen ChatUI', doc="Title of the app.")


class ExplorerUI(UI):
    """
    ExplorerUI provides a high-level entrypoint to start chatting with your data
    in split UI allowing users to load tables, explore them using Graphic Walker,
    and then interrogate the data via a chat interface.

    This high-level wrapper allows providing the data sources you will
    be chatting with and then configures the assistant and agents.


    :Example:

    .. code-block:: python

        import lumen.ai as lmai

        lmai.ExplorerUI(data='~/data.csv').servable()
    """

    chat_ui_position = param.Selector(default='left', objects=['left', 'right'], doc="""
        The position of the chat interface panel relative to the exploration area.""")

    title = param.String(default='Lumen Explorer', doc="Title of the app.")

    def __init__(
        self,
        data: DataT | list[DataT] | None = None,
        **params
    ):
        super().__init__(data=data, **params)
        cb = self.interface.callback
        self._coordinator.render_output = False
        self.interface.callback = self._wrap_callback(cb)
        self._explorations = Tabs(
            sizing_mode='stretch_both',
            closable=True,
            tabs_location="left" if self.chat_ui_position == "right" else "right",
            stylesheets=[
                """
                :host(.bk-left) .bk-header .bk-tab {
                    padding-left: 5px;
                    padding-right: 2px;
                    text-align: left;
                    max-width: 50px;
                    overflow: hidden;
                    text-overflow: clip;
                    white-space: nowrap;
                    transition: max-width 0.3s ease;
                }

                :host(.bk-left) .bk-header .bk-tab:hover {
                    max-width: 300px;
                }

                /* For right-side tabs */
                .bk-header .bk-tab {
                    max-width: 50px;
                    overflow: hidden;
                    text-overflow: clip;
                    white-space: nowrap;
                    transition: max-width 0.3s ease;
                }

                .bk-header .bk-tab:hover {
                    max-width: 300px;
                }
                """
            ]
        )
        self._explorations.param.watch(self._cleanup_explorations, ['objects'])
        self._explorations.param.watch(self._set_context, ['active'])
        self._global_notebook_export = FileDownload(
            icon="notebook",
            icon_size="1.5em",
            button_type="primary",
            callback=self._global_export_notebook,
            filename=f"{self.title.replace(' ', '_')}.ipynb",
            stylesheets=['.bk-btn a { padding: 0 6px; }'],
            styles={'position': 'absolute', 'right': '0px', 'top': '-1px', 'z-index': '100'},
        )
        self._exports.visible = False
        self._titles = []
        self._contexts = []
        self._root_conversation = self.interface.objects
        self._conversations = []

        self._explorations_intro = Markdown(
            EXPLORATIONS_INTRO,
            margin=(0, 0, 10, 10),
            sizing_mode='stretch_width',
            visible=self._explorations.param["objects"].rx.bool().rx.not_()
        )
        self._output = Tabs(
            ('Overview', self._table_explorer()),
            ('Explorations', Column(self._explorations_intro, self._explorations)),
            design=Material
        )
        output = Column(self._global_notebook_export, self._output, styles={'overflow-x': 'auto', 'overflow-y': 'auto'}, sizing_mode='stretch_both')
        chat = Column(self._exports, self._coordinator)
        self._split = SplitJS(
            left=chat,
            right=output,
            invert=self.chat_ui_position != 'left',
            sizing_mode='stretch_both'
        )
        self._main = Column(self._split)
        self._idle = asyncio.Event()
        self._idle.set()
        self._last_synced = None
        self._output.param.watch(self._update_conversation, 'active')

    def _destroy(self, session_context):
        """
        Cleanup on session destroy
        """
        for c in self._contexts:
            c.cleanup()

    def _global_export_notebook(self):
        cells, extensions = [], []
        for i, objects in enumerate(self._conversations):
            title = self._titles[i]
            header = make_md_cell(f'## {title}')
            cells.append(header)
            msg_cells, msg_exts = render_cells(objects)
            cells += msg_cells
            for ext in msg_exts:
                if ext not in extensions:
                    extensions.append(ext)
        preamble = make_preamble(self.notebook_preamble, extensions=extensions)
        return StringIO(write_notebook(preamble+cells))

    async def _update_conversation(self, event=None, tab=None):
        active = self._explorations.active
        if len(self._conversations) <= active:
            return

        if tab is None:
            # When user switches tabs and coordinator is running
            # wait to switch the conversation context
            await self._idle.wait()
            if self._explorations.active != active:
                # If active tab has changed while we waited
                # skip the sync
                return
            # If conversation was already updated, resync conversation
            if self._last_synced == active:
                if self._conversations:
                    self._conversations[active] = self.interface.objects
                else:
                    self._root_conversation = self.interface.objects
        if (event.new if event is not None else tab):
            # Explorations Tab
            if self._conversations and active < len(self._conversations):
                conversation = self._conversations[active]
            else:
                conversation = self._snapshot_messages()
            self._exports.visible = True
        else:
            # Overview tab
            self._exports.visible = False
            conversation = self._root_conversation
            # We must mark the last synced as None to ensure
            # we do not resync with the Overview tab conservation
            active = None
        self._set_conversation(conversation)
        self.interface._chat_log.scroll_to_latest()
        self._last_synced = active

    def _set_conversation(self, conversation: list[Any]):
        feed = self.interface._chat_log
        with edit_readonly(feed):
            feed.visible_range = None
        self.interface.objects = conversation

    async def _cleanup_explorations(self, event):
        if len(event.new) <= len(event.old):
            return
        for i, (old, new) in enumerate(zip(event.old, event.new)):
            if old is new:
                continue
            self._contexts.pop(i)
            self._conversations.pop(i)
            self._titles.pop(i)
            if i == self._last_synced:
                await self._update_conversation(tab=1)
            break

    async def _set_context(self, event=None, old=None, new=None):
        active = event.new if new is None else new
        if len(self._conversations) == 0:
            return
        await self._idle.wait()
        if active != self._explorations.active:
            return
        if self._last_synced == active:
            self._conversations[active] = self.interface.objects
        else:
            self._conversations[event.old if old is None else old] = self._snapshot_messages()
        conversation = self._conversations[active]
        with hold():
            self._set_conversation(conversation)
            self._notebook_export.param.update(
                filename = f"{self._titles[active].replace(' ', '_')}.ipynb"
            )
            self._exports.visible = True
        self.interface._chat_log.scroll_to_latest()
        self._last_synced = active

    def _snapshot_messages(self, new=False):
        to = -3 if new else None
        messages = []
        for msg in self.interface.objects[:to]:
            if isinstance(msg, ChatMessage):
                avatar = msg.avatar
                if isinstance(avatar, SVG) and avatar.object is PLACEHOLDER_SVG:
                    continue
            messages.append(msg)
        return messages

    def _table_explorer(self):
        from panel_gwalker import GraphicWalker

        table_select = MultiChoice(
            placeholder="Select table(s)", sizing_mode='stretch_width',
            max_height=200, max_items=5
        )
        explore_button = Button(
            name='Explore table(s)', icon='chart-bar', button_type='primary', align='center',
            disabled=table_select.param.value.rx().rx.not_()
        )
        input_row = Row(table_select, explore_button)

        source_map = {}
        def update_source_map(_, __, sources, init=False):
            selected = list(table_select.value)
            deduplicate = len(sources) > 1
            new = {}
            for source in sources:
                tables = source.get_tables()
                for t in tables:
                    if deduplicate:
                        t = f'{source.name}{SOURCE_TABLE_SEPARATOR}{t}'
                    if t.split(SOURCE_TABLE_SEPARATOR, maxsplit=1)[-1] not in source_map and not init and not len(selected) > table_select.max_items and state.loaded:
                        selected.append(t)
                    new[t] = source
            source_map.clear()
            source_map.update(new)
            selected = selected if len(selected) == 1 else []
            table_select.param.update(options=list(source_map), value=selected)
            input_row.visible = bool(source_map)
        memory.on_change('sources', update_source_map)
        update_source_map(None, None, memory['sources'], init=True)

        def explore_table_if_single(event):
            """
            If only one table is uploaded, help the user load it
            without requiring them to click twice. This step
            only triggers when the Upload in the Overview tab is used,
            i.e. does not trigger with uploads through the SourceAgent
            """
            if len(table_select.options) == 1:
                explore_button.param.trigger("value")

        controls = SourceControls(select_existing=False, cancellable=False, clear_uploads=True, multiple=True, name='Upload')
        controls.param.watch(explore_table_if_single, "add")
        tabs = Tabs(controls, dynamic=True, sizing_mode='stretch_both', design=Material)

        @param.depends(explore_button, watch=True)
        def get_explorers(load):
            if not load:
                return
            with explore_button.param.update(loading=True), self.interface.param.update(loading=True):
                explorers = []
                for table in table_select.value:
                    source = source_map[table]
                    if SOURCE_TABLE_SEPARATOR in table:
                        _, table = table.split(SOURCE_TABLE_SEPARATOR, maxsplit=1)
                    pipeline = Pipeline(
                        source=source, table=table, sql_transforms=[SQLLimit(limit=100_000, read=source.dialect)]
                    )
                    table_label = f"{table[:25]}..." if len(table) > 25 else table
                    walker = GraphicWalker(
                        pipeline.param.data, sizing_mode='stretch_both', min_height=800,
                        kernel_computation=True, name=table_label, tab='data'
                    )
                    explorers.append(walker)

                tabs.objects = explorers + [controls]
                table_select.value = []

        return Column(
            input_row,
            tabs,
            sizing_mode='stretch_both',
        )

    async def _add_exploration(self, title: str, memory: _Memory):
        n = len(self._explorations)
        active = self._explorations.active
        self._titles.append(title)
        self._contexts.append(memory)
        self.interface.objects = conversation = list(self.interface.objects)
        self._conversations.append(conversation)
        tab_title = f"{title[:25]}..." if len(title) > 25 else title
        self._explorations.append((tab_title, Column(name=title, sizing_mode='stretch_both', loading=True)))
        self._notebook_export.filename = f"{title.replace(' ', '_')}.ipynb"
        if n:
            self._conversations[active] = self._snapshot_messages(new=True)
        else:
            self._root_conversation = self._snapshot_messages(new=True)
        self._last_synced = n
        self._explorations.active = n
        self._output.active = 1
        await self._update_conversation(tab=1)

    def _add_outputs(self, exploration: Column, outputs: list[LumenOutput] | str, memory: _Memory):
        if "sql" in memory:
            sql = memory.rx("sql")
            sql_pane = Markdown(
                param.rx('```sql\n{sql}\n```').format(sql=sql),
                margin=(-15, 0, 0, 0), sizing_mode='stretch_width', name='SQL'
            )
            if sql.count('\n') > 10:
                sql_pane = Column(
                    sql_pane, max_height=325, scroll='y-auto', name='SQL'
                )
            if len(exploration) and exploration[0].name == 'SQL':
                exploration[0] = sql_pane
            else:
                exploration.insert(0, sql_pane)

        content = []
        if exploration.loading:
            from panel_gwalker import GraphicWalker
            pipeline = memory['pipeline']
            content.append(
                ('Overview', GraphicWalker(
                    pipeline.param.data,
                    kernel_computation=True,
                    tab='data',
                    sizing_mode='stretch_both'
                ))
            )
        for out in outputs:
            title = out.title or type(out).__name__.replace('Output', '')
            if len(title) > 25:
                title = f"{title[:25]}..."
            content.append((title, ParamMethod(out.render, inplace=True, sizing_mode='stretch_both')))
        if exploration.loading:
            tabs = Tabs(*content, dynamic=True, active=len(outputs))
            exploration.append(tabs)
        else:
            tabs = exploration[-1]
            tabs.extend(content)
        tabs.active = len(tabs)-1

    def _wrap_callback(self, callback):
        async def wrapper(contents: list | str, user: str, instance: ChatInterface):
            if not self._explorations:
                prev_memory = memory
            else:
                prev_memory = self._contexts[self._explorations.active]
            new_exploration = False
            prev = index = self._explorations.active if len(self._explorations) else -1
            local_memory = prev_memory.clone()
            local_memory["outputs"] = outputs = []

            async def render_plan(_, old, new):
                nonlocal index, new_exploration
                plan = local_memory["plan"]
                if any(step.actor in ('SQLAgent', 'DbtslAgent') for step in plan.steps):
                    # Expand the sidebar when the first exploration is created
                    await self._add_exploration(plan.title, local_memory)
                    index = len(self._explorations)-1
                    new_exploration = True

            def sync_available_sources_memory(_, __, sources):
                """
                For cases when the user uploads a dataset through SourceAgent
                this will update the available_sources in the global memory
                so that the overview explorer can access it
                """
                memory["sources"] += [
                    source for source in sources if source not in memory["sources"]
                ]

            local_memory.on_change('plan', render_plan)
            local_memory.on_change('sources', sync_available_sources_memory)

            async def render_output(_, old, new):
                added = [out for out in new if out not in old]
                exploration = self._explorations[index]
                self._add_outputs(exploration, added, local_memory)
                exploration.loading = False
                outputs[:] = new
                if len(self._explorations) == 1:
                    self._split.param.update(
                        collapsed=False,
                        sizes=self._split.expanded_sizes,
                    )
            local_memory.on_change('outputs', render_output)

            # Remove exploration on error if no outputs have been
            # added yet and we launched a new exploration
            async def remove_output(_, __, ___):
                nonlocal index, new_exploration
                if "__error__" in local_memory:
                    del memory['__error__']
                if outputs or not new_exploration:
                    return
                # Merge the current conversation with the conversation
                # that initiated the new exploration
                conversation = self._conversations.pop()
                if prev == -1:
                    self._root_conversation = conversation
                else:
                    self._conversations[prev] = conversation
                self._last_synced = prev
                self._contexts.pop()
                self._titles.pop()
                with hold():
                    self._explorations.active = prev
                    await self._set_context(old=index, new=prev)
                    self._explorations.pop(-1)
                    if len(self._titles) == 0:
                        self._output.active = 0
                index -= 1
                new_exploration = False
            local_memory.on_change('__error__', remove_output)

            try:
                self._idle.clear()
                with self._coordinator.param.update(memory=local_memory):
                    await callback(contents, user, instance)
            finally:
                if self._conversations:
                    self._conversations[index] = self.interface.objects
                else:
                    self._root_conversation = self.interface.objects
                self._idle.set()
                local_memory.remove_on_change('plan', render_plan)
                local_memory.remove_on_change('__error__', remove_output)
                if not new_exploration:
                    prev_memory.update(local_memory)
        return wrapper



from __future__ import annotations

import weakref

from collections import defaultdict
from functools import partial

import param

from panel.io.state import state

from ..config import SessionCache


class _Memory(SessionCache):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._session_callbacks = weakref.WeakKeyDictionary()
        self._global_callbacks = defaultdict(list)
        self._session_rx = weakref.WeakKeyDictionary()
        self._global_rx = {}

    @property
    def _callbacks(self):
        if state.curdoc:
            if state.curdoc not in self._session_callbacks:
                self._session_callbacks[state.curdoc] = defaultdict(list)
            return self._session_callbacks[state.curdoc]
        return self._global_callbacks

    @property
    def _rx(self):
        if state.curdoc:
            if state.curdoc not in self._session_rx:
                self._session_rx[state.curdoc] = {}
            return self._session_rx[state.curdoc]
        return self._global_rx

    def __setitem__(self, key, new):
        if key in self:
            old = self[key]
        else:
            old = None
        super().__setitem__(key, new)
        self._trigger_update(key, old, new)

    def cleanup(self):
        if state.curdoc:
            self._session_callbacks[state.curdoc].clear()
            self._session_rx[state.curdoc].clear()
        else:
            self._global_callbacks.clear()
            self._global_rx.clear()

    def on_change(self, key, callback):
        self._callbacks[key].append(callback)

    def remove_on_change(self, key, callback):
        self._callbacks[key].remove(callback)

    def rx(self, key):
        if key in self._rx:
            return self._rx[key]
        self._rx[key] = rxp = param.rx(self[key])
        return rxp

    def trigger(self, key):
        self._trigger_update(key, self[key], self[key])

    def _trigger_update(self, key, old, new):
        for cb in self._callbacks[key]:
            if param.parameterized.iscoroutinefunction(cb):
                param.parameterized.async_executor(partial(cb, key, old, new))
            else:
                cb(key, old, new)
        if key in self._rx:
            self._rx[key].rx.value = new


memory = _Memory()



from __future__ import annotations

import argparse
import inspect
import json
import logging
import os
import sys
import traceback

from textwrap import dedent

import bokeh.command.subcommands.serve

from bokeh.application.handlers.code import CodeHandler
from bokeh.command.util import die
from panel.command import Serve, transform_cmds
from panel.io.application import Application

try:
    from ..ai.config import THIS_DIR
except ImportError as e:
    print(f'You need to install lumen-ai with "lumen[ai]": {e}')
    sys.exit(1)

from ..ai import agents as lumen_agents, llm as lumen_llms  # Aliased here
from ..ai.utils import parse_huggingface_url, render_template

CMD_DIR = THIS_DIR / ".." / "command"

LLM_PROVIDERS = {
    'openai': 'OpenAI',
    'anthropic': 'AnthropicAI',
    'mistral': 'MistralAI',
    'azure-openai': 'AzureOpenAI',
    'azure-mistral': 'AzureMistralAI',
    "ai-navigator": "AINavigator",
    'llama-cpp': 'LlamaCpp',
}


class LLMConfig:
    """Configuration handler for LLM providers"""

    PROVIDER_ENV_VARS = {
        "openai": "OPENAI_API_KEY",
        "anthropic": "ANTHROPIC_API_KEY",
        "mistral": "MISTRAL_API_KEY",
        "azure-mistral": "AZUREAI_ENDPOINT_KEY",
        "azure-openai": "AZUREAI_ENDPOINT_KEY",
    }

    @classmethod
    def detect_provider(cls) -> str | None:
        """Detect available LLM provider based on environment variables"""
        for provider, env_var in cls.PROVIDER_ENV_VARS.items():
            if env_var and os.environ.get(env_var):
                return provider
        return None


class LumenAIServe(Serve):
    """Extended Serve command that handles both Panel/Bokeh and Lumen AI arguments"""

    def __init__(self, parser: argparse.ArgumentParser) -> None:
        super().__init__(parser=parser)
        self.add_lumen_arguments(parser)

    def add_lumen_arguments(self, parser: argparse.ArgumentParser) -> None:
        """Add Lumen AI specific arguments to the parser"""
        group = parser.add_argument_group("Lumen AI Configuration")
        group.add_argument(
            "--provider",
            choices=list(LLM_PROVIDERS.keys()),
            help="LLM provider (auto-detected from environment variables if not specified)",
        )
        group.add_argument("--api-key", help="API key for the LLM provider")
        group.add_argument("--provider-endpoint", help="Custom endpoint for the LLM provider")
        group.add_argument("--validation-mode", help="Validation mode for the LLM")
        group.add_argument("--temperature", type=float, help="Temperature for the LLM")
        group.add_argument("--agents", nargs="+", help="Additional agents to include")
        group.add_argument(
            "--model-kwargs",
            type=str,
            help="JSON string of model keyword arguments for the LLM. Example: --model-kwargs '{\"default\": {\"repo\": \"abcdef\"}}'",
        )
        group.add_argument(
            "--llm-model-url",
            type=str,
            help="""
            Huggingface URL to the GGUF file and model kwargs as query params.
            Example --llm-model-url 'https://huggingface.co/RE/PO/blob/main/FILE.gguf?chat_format=chat_format'
            """,
        )

    def invoke(self, args: argparse.Namespace) -> bool:
        """Override invoke to handle both sets of arguments"""
        provider = args.provider
        api_key = args.api_key
        endpoint = args.provider_endpoint
        mode = args.validation_mode
        temperature = args.temperature
        agents = args.agents
        log_level = args.log_level

        llm_model_url = args.llm_model_url
        if llm_model_url and provider and provider != "llama":
            raise ValueError(
                f"Cannot specify both --llm-model-url and --provider {provider!r}. "
                f"Use --llm-model-url to load a model from HuggingFace."
            )
        elif llm_model_url:
            provider = "llama"
        elif not provider:
            provider = LLMConfig.detect_provider()

        try:
            provider_cls = getattr(lumen_llms, LLM_PROVIDERS[provider])
        except (KeyError, AttributeError):
            raise ValueError(
                f"Could not find LLM Provider {provider!r}, valid providers include: {list(LLM_PROVIDERS)}."
            )

        if provider is None:
            raise RuntimeError(
                "It looks like a Language Model provider isn't set up yet.\n"
                "You have a few options to resolve this:\n\n"
                "- Set environment variables with an API key: For example, OPENAI_API_KEY or ANTHROPIC_API_KEY.\n"
                "- Specify a provider and API key directly: For example, set `--provider openai` with your API key via --api-key.\n"
                "- Custom endpoint: If using an OpenAI-compatible API, set --provider openai and define the --provider-endpoint.\n\n"
                "If you still need assistance visit the docs: https://lumen.holoviz.org/lumen_ai/how_to/llm/index.html"
            )

        model_kwargs = None
        if args.model_kwargs or llm_model_url:
            model_kwargs = {}
            if args.model_kwargs:
                try:
                    model_kwargs = json.loads(args.model_kwargs)
                except json.JSONDecodeError as e:
                    die(f"Invalid JSON format for --model-kwargs: {e}\n"
                        f"Ensure the argument is properly escaped. Example: --model-kwargs '{{\"key\": \"value\"}}'")

            # Handle huggingface URL if provided
            if llm_model_url:
                try:
                    repo, model_file, query_kwargs = parse_huggingface_url(llm_model_url)
                    # Initialize default dict if not exists
                    if 'default' not in model_kwargs:
                        model_kwargs['default'] = {}
                    model_kwargs['default'].update({
                        'repo': repo,
                        'model_file': model_file,
                        **query_kwargs
                    })
                except ValueError as e:
                    die(str(e))
        provider_cls.warmup(model_kwargs)

        agent_classes = [
            (name, cls) for name, cls in inspect.getmembers(lumen_agents, inspect.isclass)
            if issubclass(cls, lumen_agents.Agent) and cls is not lumen_agents.Agent
        ]
        agent_class_names = {name.lower(): name for name, cls in agent_classes}

        if agents:
            # Adjust agent names to match the class names, case-insensitively
            agents = [
                agent_class_names.get(agent.lower()) or
                agent_class_names.get(f"{agent.lower()}agent") or
                agent
                for agent in agents
            ]

        def build_single_handler_applications(
            paths: list[str], argvs: dict[str, list[str]] | None = None
        ) -> dict[str, Application]:
            kwargs = dict(
                provider=provider,
                api_key=api_key,
                temperature=temperature,
                endpoint=endpoint,
                mode=mode,
                agents=agents,
                log_level=log_level,
                model_kwargs=model_kwargs,
            )
            kwargs = {k: v for k, v in kwargs.items() if v is not None}
            handler = AIHandler(paths, **kwargs)
            if handler.failed:
                raise RuntimeError(
                    f"Error loading {paths}:\n\n{handler.error}\n{handler.error_detail}"
                )
            return {"/lumen_ai": Application(handler)}

        bokeh.command.subcommands.serve.build_single_handler_applications = (
            build_single_handler_applications
        )

        return super().invoke(args)


class AIHandler(CodeHandler):
    """Handler for Lumen AI applications"""

    def __init__(
        self,
        tables: list[str],
        provider: str | None = None,
        api_key: str | None = None,
        temperature: float | None = None,
        endpoint: str | None = None,
        mode: str | None = None,
        agents: list[str] | None = None,
        log_level: str = "INFO",
        model_kwargs: dict | None = None,
        **kwargs,
    ) -> None:
        source = self._build_source_code(
            tables=tables,
            provider=provider,
            api_key=api_key,
            temperature=temperature,
            endpoint=endpoint,
            mode=mode,
            agents=agents,
            log_level=log_level,
            model_kwargs=model_kwargs,
        )
        super().__init__(filename="lumen_ai.py", source=source, **kwargs)

    def _build_source_code(self, tables: list[str], **config) -> str:
        """Build source code with configuration"""
        context = {
            "llm_provider": LLM_PROVIDERS[config['provider']],
            "tables": [repr(t) for t in tables],
            "api_key": config.get("api_key"),
            "endpoint": config.get("endpoint"),
            "mode": config.get("mode"),
            "agents": config.get("agents"),
            "log_level": config["log_level"],
            "temperature": config.get("temperature"),
            "model_kwargs": config.get('model_kwargs') or {},
        }
        context = {k: v for k, v in context.items() if v is not None}

        source = render_template(
            CMD_DIR / "app.py.jinja2", relative_to=CMD_DIR, **context
        ).replace("\n\n", "\n").strip()
        return source


def main(args=None):
    parser = argparse.ArgumentParser(
        prog="lumen-ai",
        description=dedent("""\
            Lumen AI - Launch Lumen AI applications with customizable LLM configuration.
            To start the application without any data, simply run 'lumen-ai' with no additional arguments.

            First time running Lumen AI take a look at getting started documentation:
            https://lumen.holoviz.org/lumen_ai/getting_started/

            Found a Bug or Have a Feature Request?
            Open an issue at: https://github.com/holoviz/lumen/issues

            Have a Question?
            Ask on our Discord chat server: https://discord.gg/rb6gPXbdAr

            Need Help?
            Ask a question on our forum: https://discourse.holoviz.org
        """),
        formatter_class=argparse.RawTextHelpFormatter,
        epilog="See '<command> --help' to read about a specific subcommand.",
    )

    parser.add_argument("-v", "--version", action="version", version="Lumen AI 1.0.0")

    subs = parser.add_subparsers(help="Sub-commands", dest="command")

    serve_parser = subs.add_parser(
        Serve.name, help="Run a bokeh server to serve the Lumen AI application."
    )
    serve_command = LumenAIServe(parser=serve_parser)
    serve_parser.set_defaults(invoke=serve_command.invoke)

    if len(sys.argv) > 1 and sys.argv[1] in ("--help", "-h"):
        args = parser.parse_args(sys.argv[1:])
        args.invoke(args)
        sys.exit()

    if len(sys.argv) == 1:
        sys.argv.extend(["serve", "no_data"])

    sys.argv = transform_cmds(sys.argv)
    args = parser.parse_args(sys.argv[1:])

    if not hasattr(args, "invoke"):
        parser.print_help()
        sys.exit(1)

    try:
        ret = args.invoke(args)
    except Exception as e:
        levels = logging.getLevelNamesMapping()
        if levels.get((args.log_level or 'warn').upper(), 30) < 20:
            traceback.print_exc()
        die("ERROR: " + str(e))

    if ret is False:
        sys.exit(1)
    elif ret is not True and isinstance(ret, int) and ret != 0:
        sys.exit(ret)


if __name__ == "__main__":
    main()



import datetime
import json
import sqlite3
import uuid

from abc import abstractmethod
from functools import wraps
from pathlib import Path
from typing import Any, Literal

import param

from pydantic import BaseModel

from .utils import log_debug


class Message(BaseModel):
    role: str
    content: str


class Invocation(BaseModel):
    input_id: int
    prompt: str
    messages: list[Message]
    response: str | None
    kwargs: dict[str, Any]
    invocation_id: str

    def serialize(self) -> list[dict[str, Any]]:
        """Serialize messages into a list of dictionaries."""
        return [
            {"role": message.role, "content": message.content}
            for message in self.messages
        ]


class Session(BaseModel):
    session_id: str
    invocations: list[Invocation]


class Interceptor(param.Parameterized):

    db_path = param.String(
        default="messages.db", doc="Path to the SQLite database file"
    )

    def __init__(self, **params):
        super().__init__(**params)
        needs_init = not Path(self.db_path).exists()
        self.conn = self._create_connection()
        if needs_init:
            self.init_db()
        self._client = self._original_create = self._original_create_response = None
        self._last_invocation_id = None
        self.session_id = self._generate_session_id()

    def _create_connection(self) -> sqlite3.Connection:
        """Create and return a database connection."""
        return sqlite3.connect(self.db_path)

    def _generate_session_id(self) -> str:
        """Generate a unique session ID."""
        first_message_timestamp = datetime.datetime.now().strftime("%Y%m%dT%H%M%S")
        return f"session_{first_message_timestamp}"

    def _dump_response_model(self, response_model: BaseModel) -> str:
        """Dump the response model to a JSON string."""
        return json.dumps(response_model.model_dump())

    @abstractmethod
    def patch_client(self, client) -> None:
        """
        Patch the LLM client's create method to store messages and arguments in the database.

        Args:
            client: The LLM client instance to patch.
        """

    def init_db(self) -> None:
        """Initialize the database by creating necessary tables if they don't exist."""
        cursor = self.conn.cursor()
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS sessions (
                session_id TEXT PRIMARY KEY,
                created_at DATETIME DEFAULT CURRENT_TIMESTAMP
            )
            """
        )
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS invocations (
                invocation_id TEXT PRIMARY KEY,
                session_id TEXT,
                messages TEXT,
                message_kwargs TEXT,
                timestamp DATETIME DEFAULT CURRENT_TIMESTAMP,
                FOREIGN KEY (session_id) REFERENCES sessions(session_id)
            )
            """
        )
        self.conn.commit()
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS responses (
                response_id INTEGER PRIMARY KEY AUTOINCREMENT,
                invocation_id INTEGER,
                content TEXT,
                timestamp DATETIME DEFAULT CURRENT_TIMESTAMP,
                FOREIGN KEY (invocation_id) REFERENCES invocations(invocation_id)
            )
            """
        )

    def reset_db(self) -> None:
        """Reset the database by deleting all tables."""
        cursor = self.conn.cursor()
        cursor.execute("DROP TABLE IF EXISTS invocations")
        cursor.execute("DROP TABLE IF EXISTS sessions")
        self.conn.commit()
        self.init_db()

    def delete_session(self, session_id: str | None = None) -> None:
        """Delete the last session from the database."""
        cursor = self.conn.cursor()
        if session_id is None:
            session_id = self.get_session_ids()[-1]
        cursor.execute("DELETE FROM invocations WHERE session_id = ?", (session_id,))
        self.conn.commit()

    def store_invocation(
        self, messages: list[dict[str, str]], **kwargs: dict[str, Any]
    ) -> None:
        """
        Store messages and keyword arguments in the database for the current session.

        Args:
            messages: List of message dictionaries to store.
            kwargs: The keyword arguments passed to the create method.
        """
        cursor = self.conn.cursor()
        self._last_invocation_id = invocation_id = uuid.uuid4().hex

        cursor.execute(
            """
            INSERT INTO invocations (invocation_id, session_id, messages, message_kwargs)
            VALUES (?, ?, ?, ?)
            """,
            (invocation_id, self.session_id, json.dumps(messages), json.dumps(kwargs)),
        )
        self.conn.commit()

    def store_response(self, content: str) -> None:
        cursor = self.conn.cursor()
        cursor.execute(
            """
            INSERT INTO responses (invocation_id, content, timestamp)
            VALUES (?, ?, ?)
            """,
            (self._last_invocation_id, content, datetime.datetime.now()),
        )
        self.conn.commit()

    def get_session(self, session_id: str | None = None) -> Session:
        """
        Retrieve the session invocations of inputs from the last session, or a specific session if provided.

        Args:
            session_id: The session ID to retrieve invocations from. If not provided, the last session is used.

        Returns:
            A Session object containing invocations for the session.
        """
        cursor = self.conn.cursor()

        if session_id is None:
            cursor.execute(
                "SELECT session_id FROM invocations ORDER BY timestamp DESC LIMIT 1"
            )
            try:
                session_id = cursor.fetchone()[0]
            except TypeError:
                return Session(session_id="", invocations=[])

        cursor.execute(
            """
            SELECT
                invocations.invocation_id,
                messages,
                message_kwargs,
                responses.content AS response
            FROM invocations
            LEFT JOIN responses ON invocations.invocation_id = responses.invocation_id
            WHERE invocations.session_id = ?
            ORDER BY invocations.timestamp ASC
            """,
            (session_id,),
        )
        invocation_data = cursor.fetchall()

        input_id = -1
        invocations = []
        prev_user_content = None
        for invocation_id, invocation_messages, message_kwargs, response in invocation_data:
            messages = []
            for message in json.loads(invocation_messages):
                if message["role"] == "user":
                    user_content = message["content"]
                    if prev_user_content != user_content:
                        prev_user_content = user_content
                        input_id += 1
                messages.append(Message(role=message["role"], content=message["content"]))

            invocations.append(
                Invocation(
                    prompt=user_content,
                    invocation_id=invocation_id,
                    input_id=input_id,
                    messages=messages,
                    kwargs=json.loads(message_kwargs),
                    response=response,
                )
            )

        return Session(session_id=session_id, invocations=invocations)

    def get_all_sessions(self) -> dict[str, Session]:
        """
        Retrieve the invocations of messages from all sessions.

        Returns:
            A dictionary containing session_id as keys and the corresponding
                Session object for each session.
        """
        all_sessions = {}
        for session_id in self.get_session_ids():
            all_sessions[session_id] = self.get_session(session_id)

        return all_sessions

    def get_session_ids(self) -> list[str]:
        cursor = self.conn.cursor()
        cursor.execute("SELECT DISTINCT session_id FROM invocations")
        return [row[0] for row in cursor.fetchall()]

    def unpatch(self) -> None:
        """Close the database connection and revert the client create."""
        if self._original_create_response is not None:
            self._client.chat.completions.create = self._original_create_response
        if self._original_create is not None:
            self._client.chat.completions.create = self._original_create
        self.conn.close()

    def __del__(self) -> None:
        """Close the database connection and revert the client create when the object is deleted."""
        self.unpatch()


class OpenAIInterceptor(Interceptor):

    def patch_client(
        self, client, mode: Literal["store_all", "store_inputs"] = "store_all"
    ) -> None:
        """
        Patch the OpenAI client's create method to store messages and arguments in the database.

        Args:
            client: The OpenAI client instance to patch.
            mode: The mode to patch the client in.
                If "store_all", responses are generated, and everything is stored.
                If "store_inputs", responses are generated, but only input arguments are stored.
        """
        self._client = client
        self._original_create = client.chat.completions.create

        async def stream_response(*args: Any, **kwargs: Any):
            async for chunk in await self._original_create(*args, **kwargs):
                yield chunk
            self.store_invocation(**kwargs)

        async def non_stream_response(*args: Any, **kwargs: Any):
            response = await self._original_create(*args, **kwargs)
            self.store_invocation(**kwargs)
            return response

        @wraps(client.chat.completions.create)
        async def patched_async_create(*args: Any, **kwargs: Any) -> Any:
            stream = kwargs.get("stream", False)
            if stream:
                return stream_response(*args, **kwargs)
            else:
                return await non_stream_response(*args, **kwargs)

        self._client.chat.completions.create = patched_async_create
        if mode == "store_all":
            self.patch_client_response(client)

    def patch_client_response(self, client) -> None:
        """
        Patch the OpenAI client's create method to store the response in the database.

        Args:
            client: The OpenAI client instance to patch.
        """
        self._client = client
        self._original_create_response = client.chat.completions.create

        async def stream_response(*args: Any, **kwargs: Any):
            content = ""
            response = await self._original_create_response(*args, **kwargs)
            async for chunk in response:
                yield chunk
                if hasattr(chunk, "choices") and chunk.choices:
                    delta = chunk.choices[0].delta
                    if delta.content is not None:
                        content += delta.content

            # last chunk contains all the content
            if not content and isinstance(chunk, BaseModel):
                content = self._dump_response_model(chunk)

            if content:
                self.store_response(content)
            else:
                log_debug("Could not intercept a response")

        async def non_stream_response(*args: Any, **kwargs: Any):
            content = ""
            response = await self._original_create_response(*args, **kwargs)
            if hasattr(response, "choices"):
                content = response.choices[0].message.content

            if not content and isinstance(response, BaseModel):
                content = self._dump_response_model(response)

            if content:
                self.store_response(content)
            else:
                log_debug("Could not intercept a response.")

            return response

        @wraps(client.chat.completions.create)
        async def patched_create_response(*args: Any, **kwargs: Any) -> Any:
            stream = kwargs.get("stream", False)
            if stream:
                return stream_response(*args, **kwargs)
            else:
                return await non_stream_response(*args, **kwargs)

        client.chat.completions.create = patched_create_response



from __future__ import annotations

from typing import Literal

from instructor.dsl.partial import PartialLiteralMixin
from pydantic import BaseModel, Field, create_model
from pydantic.fields import FieldInfo

from .config import SOURCE_TABLE_SEPARATOR


class PartialBaseModel(BaseModel, PartialLiteralMixin):
    ...


class YesNo(BaseModel):

    yes: bool = Field(description="True if yes, otherwise False.")


class ThinkingYesNo(YesNo):

    chain_of_thought: str = Field(
        description="Explain concisely how you arrived at your answer in under a sentence, or even a phrase.")


class Sql(BaseModel):

    chain_of_thought: str = Field(
        description="""
        You are a world-class SQL expert, and your fame is on the line so don't mess up.
        If it's simple, give at max one sentence about using `SELECT * FROM ...`.
        However, if applicable, be sure to carefully study the schema, discuss the values in the columns,
        and whether you need to wrangle the data before you can use it, before finally writing a correct and valid
        SQL query that fully answers the user's query. If using CTEs, comment on the purpose of each.
        Everything should be made as simple as possible, but no simpler and be concise.
        """
    )

    expr_slug: str = Field(
        description="""
        Give the SQL expression a concise, but descriptive, slug that includes whatever transforms were applied to it,
        e.g. top_5_athletes. The slug must be unique, i.e. should not match other existing table names or slugs.
        """
    )

    query: str = Field(description="""
        Correct, valid SQL query that answers the user's query and is based on
        the chain of thought; do NOT add extraneous comments.""")


class VegaLiteSpec(BaseModel):

    chain_of_thought: str = Field(
        description="Explain how you will use the data to create a vegalite plot, and address any previous issues you encountered."
    )

    json_spec: str = Field(description="A vega-lite JSON specification based on the user input and chain of thought. Do not include description")



class RetrySpec(BaseModel):

    chain_of_thought: str = Field(
        description="Explain why the previous spec failed to address the user query and what you will do differently this time to ensure it is correct."
    )

    corrected_spec: str = Field(
        description="The corrected version of the previous spec without any additional comments, additions or code examples."
    )


def make_plan_models(agents: list[str], tools: list[str]):
    # TODO: make this inherit from PartialBaseModel
    step = create_model(
        "Step",
        actor=(Literal[tuple(agents+tools)], FieldInfo(description="The name of the actor to assign a task to.")),
        instruction=(str, FieldInfo(description="Instructions to the actor to assist in the task, and whether rendering is required.")),
        title=(str, FieldInfo(description="Short title of the task to be performed; up to three words.")),
        render_output=(bool, FieldInfo(description="Whether the output of the actor should be rendered. If the user wants to see the table, and the actor is SQL, then this should be `True`.")),
    )
    reasoning = create_model(
        'Reasoning',
        chain_of_thought=(
            str,
            FieldInfo(
                description="""
                    Briefly summarize the user's goal and categorize the question type:
                    high-level, data-focused, or other. Identify the most relevant and compatible actors,
                    explaining their requirements, and what you already have satisfied. If there were previous failures, discuss them.
                    """
            ),
        ),
    )
    plan = create_model(
        "Plan",
        title=(
            str, FieldInfo(description="A title that describes this plan, up to three words.")
        ),
        steps=(
            list[step],
            FieldInfo(
                description="A list of steps to perform that will solve user query. Ensure you include ALL the steps needed to solve the task, matching the chain of thought."
            )
        )
    )
    return reasoning, plan


def make_columns_selection(table_slugs: list[str], **context):

    class TableColumnsIndices(PartialBaseModel):

        table_slug: Literal[tuple(table_slugs)] = Field(
            description=f"The table slug, i.e. '<source>{SOURCE_TABLE_SEPARATOR}<table>'"
        )

        column_indices: list[int] = Field(
            description="A list of 0-based column indices for the table specified by `table_slug`. This indicates which columns should be included in the output."
        )

    class ColumnsSelection(PartialBaseModel):
        """
        Model for selecting a subset of columns from tables.
        """
        chain_of_thought: str = Field(
            description="""
            Break down the user query into parts, and try to map out the columns to
            the parts of the query that they are relevant to. Select more
            than you need, and then you can filter them out later.
            """
        )

        tables_columns_indices: list[TableColumnsIndices] = Field(
            description="""
            The list of table slugs, i.e. '<source>{SOURCE_TABLE_SEPARATOR}<table>',
            and their respective columns based on your own chain_of_thought."""
        )
    return ColumnsSelection


def make_agent_model(agent_names: list[str], primary: bool = False):
    if primary:
        description = "The agent that will provide the output the user requested, e.g. a plot or a table. This should be the FINAL step in your chain of thought."
    else:
        description = "The most relevant agent to use."
    return create_model(
        "Agent",
        chain_of_thought=(
            str,
            FieldInfo(
                description="Describe what this agent should do."
            ),
        ),
        agent_or_tool=(
            Literal[tuple(agent_names)],
            FieldInfo(default=..., description=description)
        ),
    )


def make_find_tables_model(tables):
    table_model = create_model(
        "Table",
        chain_of_thought=(str, FieldInfo(
            description="""
            Concisely consider which tables are necessary to answer the user query.
            If there are tables that provide the same info, do not include them all; instead, pick the most relevant one.
            """
        )),
        selected_tables=(list[Literal[tuple(tables)]], FieldInfo(
            description="""
            The most relevant tables based on the user query; if none are relevant,
            use the first table. At least one table must be provided.
            If a join is necessary, include all the tables that will be used in the join.
            """
        )),
       potential_join_issues=(str, FieldInfo(
           description="""
           If no join is necessary, return an empty string--else
           list potential join issues between tables:
           - Data type mismatches (int vs string, numeric precision)
           - Format differences (case, leading zeros, dates/times, timezones)
           - Semantic differences (IDs vs names, codes vs full text)
           - Quality issues (nulls, duplicates, validation rules)
           Return specific issues found in current tables, and how you plan to address them
           in the most easiest, but accurate way possible.
           """
       )),
        __base__=PartialBaseModel
    )
    return table_model


def make_iterative_selection_model(table_slugs):
    """
    Creates a model for table selection in the coordinator.
    This is separate from the SQLAgent's table selection model to focus on context gathering
    rather than query execution.
    """
    table_model = create_model(
        "IterativeTableSelection",
        chain_of_thought=(str, FieldInfo(
            description="""
            Consider which tables would provide the most relevant context for understanding the user query,
            and be specific about what columns are most relevant to the query. Please be concise.
            """
        )),
        is_done=(bool, FieldInfo(
            description="""
            Indicate whether you have sufficient context about the available data structures.
            Set to True if you understand enough about the data model to proceed with the query.
            Set to False if you need to examine additional tables to better understand the domain.
            Do not reach for perfection; if there are promising tables, but you are not 100% sure,
            still mark as done.
            """
        )),
        requires_join=(bool, FieldInfo(
            description="""
            Indicate whether a join is necessary to answer the user query.
            Set to True if a join is necessary to answer the query.
            Set to False if no join is necessary.
            """
        )),
        selected_slugs=(list[Literal[tuple(table_slugs)]], FieldInfo(
            description="""
            If is_done and a join is necessary, return a list of table names that will be used in the join.
            If is_done and no join is necessary, return a list of a single table name.
            If not is_done return a list of table names that you believe would provide
            the most relevant context for understanding the user query that wasn't already seen before.
            Focus on tables that contain key entities, relationships, or metrics mentioned in the query.
            """
        )),
        __base__=PartialBaseModel
    )
    return table_model


class DbtslQueryParams(BaseModel):
    """
    Model for dbtsl.client.query() parameters.
    """

    chain_of_thought: str = Field(
        description="""You are a world-class dbt Semantic Layer expert. Think step by step about
        what metrics are needed, what dimensions to group by, what time granularity
        to use, and any filters that should be applied; if filters are applied, include those
        filtered dimensions in group_by. If there are errors, mention how you'll address the errors.
        """
    )

    expr_slug: str = Field(
        description="""Give the query a concise, but descriptive, slug that includes the metrics
        and dimensions used, e.g. monthly_revenue_by_region. The slug must be unique."""
    )

    where: list[str] = Field(
        default_factory=list,
        description="A list of conditions to filter the results; dimensions referenced here must also be in group_by, e.g. ['metric_time__month >= date_trunc('month', '2024-09-30'::date)']"
    )

    group_by: list[str] = Field(
        default_factory=list,
        description="A list of dimensions to group by, e.g. ['metric_time__month'], must include dimensions from where."
    )

    limit: int = Field(
        default=None,
        description="The maximum number of rows to return."
    )

    metrics: list[str] = Field(
        default_factory=list,
        description="A list of metrics to include in the query, e.g. ['revenue']"
    )

    order_by: list[str] = Field(
        default_factory=list,
        description="A list of columns or expressions to order the results by, e.g. ['metric_time__month']"
    )


def make_refined_query_model(item_type_name: str = "items"):
    """
    Creates a model for refining search queries in vector lookup tools.
    """
    return create_model(
        "RefinedQuery",
        chain_of_thought=(str, Field(
            description=f"""
            Analyze the current search results for {item_type_name}. Consider whether the terms used
            in the query match the terms that might be found in {item_type_name} names and descriptions.
            Think about more specific or alternative terms that might yield better results.
            """
        )),
        refined_search_query=(str, Field(
            description=f"""
            A refined search query that would help find more relevant {item_type_name}.
            This should be a focused query with specific terms, entities, or concepts
            that might appear in relevant {item_type_name}.
            """
        )),
        __base__=PartialBaseModel
    )
